import ffilib
import uctypes
import struct

from collections import namedtuple


libc = ffilib.libc()

getpwnam_ = libc.func("P", "getpwnam", "s")


struct_passwd = namedtuple(
    "struct_passwd", ["pw_name", "pw_passwd", "pw_uid", "pw_gid", "pw_gecos", "pw_dir", "pw_shell"]
)


def getpwnam(user):
    passwd = getpwnam_(user)
    if not passwd:
        raise KeyError("getpwnam(): name not found: {}".format(user))
    passwd_fmt = "SSIISSS"
    passwd = uctypes.bytes_at(passwd, struct.calcsize(passwd_fmt))
    passwd = struct.unpack(passwd_fmt, passwd)
    return struct_passwd(*passwd)



import io

try:
    io.BytesIO
    io.BufferedWriter
except AttributeError:
    print('SKIP')
    raise SystemExit

bts = io.BytesIO()
buf = io.BufferedWriter(bts, 8)

buf.write(b"foobar")
print(bts.getvalue())
buf.write(b"foobar")
# CPython has different flushing policy, so value below is different
print(bts.getvalue())
buf.flush()
print(bts.getvalue())
buf.flush()
print(bts.getvalue())

# special case when alloc is a factor of total buffer length
bts = io.BytesIO()
buf = io.BufferedWriter(bts, 1)
buf.write(b"foo")
print(bts.getvalue())

# hashing a BufferedWriter
print(type(hash(buf)))



# test overriding builtins

import builtins

# override generic builtin
try:
    builtins.abs = lambda x: x + 1
except AttributeError:
    print("SKIP")
    raise SystemExit

print(abs(1))

# __build_class__ is handled in a special way
orig_build_class = __build_class__
builtins.__build_class__ = lambda x, y: ('class', y)
class A:
    pass
print(A)
builtins.__build_class__ = orig_build_class

# __import__ is handled in a special way
def custom_import(name, globals, locals, fromlist, level):
    print('import', name, fromlist, level)
    class M:
        a = 1
        b = 2
    return M
builtins.__import__ = custom_import
__import__('A', None, None, None, 0)
import a
import a.b
from a import a
from a.b import a, b
from .a import a
from ..a import a, b



# This test checks previously known problem values for 32-bit ports.
# It's less useful for 64-bit ports.
try:
    import uctypes
except ImportError:
    print("SKIP")
    raise SystemExit

buf = bytearray(b"12345678abcd")
struct = uctypes.struct(
    uctypes.addressof(buf),
    {"f32": uctypes.UINT32 | 0, "f64": uctypes.UINT64 | 4},
    uctypes.LITTLE_ENDIAN,
)

struct.f32 = 0x7FFFFFFF
print(buf)

struct.f32 = 0x80000000
print(buf)

struct.f32 = 0xFF010203
print(buf)

struct.f64 = 0x80000000
print(buf)

struct.f64 = 0x80000000 * 2
print(buf)

print("=")

buf = bytearray(b"12345678abcd")
struct = uctypes.struct(
    uctypes.addressof(buf),
    {"f32": uctypes.UINT32 | 0, "f64": uctypes.UINT64 | 4},
    uctypes.BIG_ENDIAN,
)

struct.f32 = 0x7FFFFFFF
print(buf)

struct.f32 = 0x80000000
print(buf)

struct.f32 = 0xFF010203
print(buf)

struct.f64 = 0x80000000
print(buf)

struct.f64 = 0x80000000 * 2
print(buf)



# test assigning to iterator within the loop
for i in range(2):
    print(i)
    i = 2

# test assigning to range parameter within the loop
# (since we optimise for loops, this needs checking, currently it fails)
n = 2
for i in range(n):
    print(i)
    n = 0



# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2017 Glenn Ruben Bakke
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE

# MicroPython controller for PowerUp 3.0 paper airplane
# https://www.poweruptoys.com/products/powerup-v3
#
# Examples is written for nrf52832, pca10040 using s132 bluetooth stack.
#
# Joystick shield pin mapping:
#  - analog stick x-direction - ADC0 - P0.02/"P2"
#  - buttons P0.13 - P0.16 / "P13", "P14", "P15", "P16"
#
# Example usage:
#
#     from powerup import PowerUp3
#     p = PowerUp3()

import time
from machine import ADC
from machine import Pin
from ubluepy import Peripheral, Scanner, constants


def bytes_to_str(bytes):
    string = ""
    for b in bytes:
        string += chr(b)
    return string


def get_device_names(scan_entries):
    dev_names = []
    for e in scan_entries:
        scan = e.getScanData()
        if scan:
            for s in scan:
                if s[0] == constants.ad_types.AD_TYPE_COMPLETE_LOCAL_NAME:
                    dev_names.append((e, bytes_to_str(s[2])))
    return dev_names


def find_device_by_name(name):
    s = Scanner()
    scan_res = s.scan(500)

    device_names = get_device_names(scan_res)
    for dev in device_names:
        if name == dev[1]:
            return dev[0]


class PowerUp3:
    def __init__(self):
        self.x_adc = ADC(1)

        self.btn_speed_up = Pin("P13", mode=Pin.IN, pull=Pin.PULL_UP)
        self.btn_speed_down = Pin("P15", mode=Pin.IN, pull=Pin.PULL_UP)
        self.btn_speed_full = Pin("P14", mode=Pin.IN, pull=Pin.PULL_UP)
        self.btn_speed_off = Pin("P16", mode=Pin.IN, pull=Pin.PULL_UP)

        self.x_mid = 0

        self.calibrate()
        self.connect()
        self.loop()

    def read_stick_x(self):
        return self.x_adc.value()

    def button_speed_up(self):
        return not bool(self.btn_speed_up.value())

    def button_speed_down(self):
        return not bool(self.btn_speed_down.value())

    def button_speed_full(self):
        return not bool(self.btn_speed_full.value())

    def button_speed_off(self):
        return not bool(self.btn_speed_off.value())

    def calibrate(self):
        self.x_mid = self.read_stick_x()

    def __str__(self):
        return "calibration x: %i, y: %i" % (self.x_mid)

    def map_chars(self):
        s = self.p.getServices()

        service_batt = s[3]
        service_control = s[4]

        self.char_batt_lvl = service_batt.getCharacteristics()[0]
        self.char_control_speed = service_control.getCharacteristics()[0]
        self.char_control_angle = service_control.getCharacteristics()[2]

    def battery_level(self):
        return int(self.char_batt_lvl.read()[0])

    def speed(self, new_speed=None):
        if new_speed is None:
            return int(self.char_control_speed.read()[0])
        else:
            self.char_control_speed.write(bytearray([new_speed]))

    def angle(self, new_angle=None):
        if new_angle is None:
            return int(self.char_control_angle.read()[0])
        else:
            self.char_control_angle.write(bytearray([new_angle]))

    def connect(self):
        dev = None

        # connect to the airplane
        while not dev:
            dev = find_device_by_name("TailorToys PowerUp")
            if dev:
                self.p = Peripheral()
                self.p.connect(dev.addr())

        # locate interesting characteristics
        self.map_chars()

    def rudder_center(self):
        if self.old_angle != 0:
            self.old_angle = 0
            self.angle(0)

    def rudder_left(self, angle):
        steps = angle // self.interval_size_left
        new_angle = 60 - steps

        if self.old_angle != new_angle:
            self.angle(new_angle)
            self.old_angle = new_angle

    def rudder_right(self, angle):
        steps = angle // self.interval_size_right
        new_angle = -steps

        if self.old_angle != new_angle:
            self.angle(new_angle)
            self.old_angle = new_angle

    def throttle(self, speed):
        if speed > 200:
            speed = 200
        elif speed < 0:
            speed = 0

        if self.old_speed != speed:
            self.speed(speed)
            self.old_speed = speed

    def loop(self):
        adc_threshold = 10
        right_threshold = self.x_mid + adc_threshold
        left_threshold = self.x_mid - adc_threshold

        self.interval_size_left = self.x_mid // 60
        self.interval_size_right = (255 - self.x_mid) // 60

        self.old_angle = 0
        self.old_speed = 0

        while True:
            time.sleep_ms(100)

            # read out new angle
            new_angle = self.read_stick_x()
            if new_angle < 256:
                if new_angle > right_threshold:
                    self.rudder_right(new_angle - self.x_mid)
                elif new_angle < left_threshold:
                    self.rudder_left(new_angle)
                else:
                    self.rudder_center()

            # read out new speed
            new_speed = self.old_speed

            if self.button_speed_up():
                new_speed += 25
            elif self.button_speed_down():
                new_speed -= 25
            elif self.button_speed_full():
                new_speed = 200
            elif self.button_speed_off():
                new_speed = 0
            else:
                pass

            self.throttle(new_speed)



# test handling of failed heap allocation with list

import micropython


class GetSlice:
    def __getitem__(self, idx):
        return idx


sl = GetSlice()[:]

# create slice in VM
l = [1, 2, 3]
micropython.heap_lock()
try:
    print(l[0:1])
except MemoryError:
    print("MemoryError: list index")
micropython.heap_unlock()

# get from list using slice
micropython.heap_lock()
try:
    l[sl]
except MemoryError:
    print("MemoryError: list get slice")
micropython.heap_unlock()

# extend list using slice subscr
l = [1, 2]
l2 = [3, 4, 5, 6, 7, 8, 9, 10]
micropython.heap_lock()
try:
    l[sl] = l2
except MemoryError:
    print("MemoryError: list extend slice")
micropython.heap_unlock()
print(l)



# Test ilistdir __del__ for VfsFat using a RAM device.
import gc

try:
    import os, vfs

    vfs.VfsFat
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMBlockDevice:
    ERASE_BLOCK_SIZE = 512

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.ERASE_BLOCK_SIZE)

    def readblocks(self, block, buf, off=0):
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            buf[i] = self.data[addr + i]

    def writeblocks(self, block, buf, off=0):
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            self.data[addr + i] = buf[i]

    def ioctl(self, op, arg):
        if op == 4:  # block count
            return len(self.data) // self.ERASE_BLOCK_SIZE
        if op == 5:  # block size
            return self.ERASE_BLOCK_SIZE
        if op == 6:  # erase block
            return 0


def test(bdev, vfs_class):
    vfs_class.mkfs(bdev)
    fs = vfs_class(bdev)
    fs.mkdir("/test_d1")
    fs.mkdir("/test_d2")
    fs.mkdir("/test_d3")

    for i in range(10):
        print(i)

        # We want to partially iterate the ilistdir iterator to leave it in an
        # open state, which will then test the finaliser when it's garbage collected.
        idir = fs.ilistdir("/")
        print(any(idir))

        # Alternate way of partially iterating the ilistdir object, modifying the
        # filesystem while it's open.
        for dname, *_ in fs.ilistdir("/"):
            fs.rmdir(dname)
            break
        fs.mkdir(dname)

        # Also create a fully drained iterator and ensure trying to reuse it
        # throws the correct exception.
        idir_emptied = fs.ilistdir("/")
        l = list(idir_emptied)
        print(len(l))
        try:
            next(idir_emptied)
        except StopIteration:
            pass

        gc.collect()
        fs.open("/test", "w").close()


try:
    bdev = RAMBlockDevice(50)
except MemoryError:
    print("SKIP")
    raise SystemExit

test(bdev, vfs.VfsFat)



# FeatherS2 MicroPython Helper Library
# 2021 Seon Rozenblum - Unexpected Maker
#
# Project home:
#   https://feathers2.io
#
# 2021-Mar-21 - v0.1 - Initial implementation

# Import required libraries
from micropython import const
from machine import Pin, SPI, ADC
import machine, time

# FeatherS2 Hardware Pin Assignments

# LDO
LDO2 = const(21)

# APA102 Dotstar pins
DOTSTAR_CLK = const(45)
DOTSTAR_DATA = const(40)

# SPI
SPI_MOSI = const(35)
SPI_MISO = const(37)
SPI_CLK = const(36)

# I2C
I2C_SDA = const(8)
I2C_SCL = const(9)

# DAC
DAC1 = const(17)
DAC2 = const(18)

# LED & Ambient Light Sensor
LED = const(13)
AMB_LIGHT = const(4)

# Helper functions


# LED & Ambient Light Sensor control
def set_led(state):
    l = Pin(LED, Pin.OUT)
    l.value(state)


def toggle_led(state):
    l = Pin(LED, Pin.OUT)
    l.value(not l.value())


# Create ADC and set attenuation and return the ambient light value from the onboard sensor
def get_amb_light():
    adc = ADC(Pin(AMB_LIGHT))
    adc.atten(ADC.ATTN_11DB)
    return adc.read()


# LDO2 power control
# When we manually turn off the second LDO we also set the DotStar DATA and CLK pins to input to
# prevent parasitic power from lighting the LED even with the LDO off, causing current use.
# The DotStar is a beautiful LED, but parasitic power makes it a terrible choice for battery use :(
def set_ldo2_power(state):
    """Set the power for the on-board Dotstar to allow no current draw when not needed."""
    # Set the power pin to the inverse of state
    ldo2 = Pin(LDO2, Pin.OUT)
    ldo2.value(state)

    if state:
        Pin(DOTSTAR_CLK, Pin.OUT)
        Pin(DOTSTAR_DATA, Pin.OUT)  # If power is on, set CLK to be output, otherwise input
    else:
        Pin(DOTSTAR_CLK, Pin.IN)
        Pin(DOTSTAR_DATA, Pin.IN)  # If power is on, set CLK to be output, otherwise input

    # A small delay to let the IO change state
    time.sleep(0.035)


# Dotstar rainbow colour wheel
def dotstar_color_wheel(wheel_pos):
    """Color wheel to allow for cycling through the rainbow of RGB colors."""
    wheel_pos = wheel_pos % 255

    if wheel_pos < 85:
        return 255 - wheel_pos * 3, 0, wheel_pos * 3
    elif wheel_pos < 170:
        wheel_pos -= 85
        return 0, wheel_pos * 3, 255 - wheel_pos * 3
    else:
        wheel_pos -= 170
        return wheel_pos * 3, 255 - wheel_pos * 3, 0


# Go into deep sleep but shut down the APA first to save power
# Use this if you want lowest deep  sleep current
def go_deepsleep(t):
    """Deep sleep helper that also powers down the on-board Dotstar."""
    set_ldo2_power(False)
    machine.deepsleep(t)



try:
    from io import StringIO
    import json
except ImportError:
    print("SKIP")
    raise SystemExit

s = StringIO()
json.dump(False, s)
print(s.getvalue())

s = StringIO()
json.dump({"a": (2, [3, None])}, s)
print(s.getvalue())

# dump to a small-int not allowed
try:
    json.dump(123, 1)
except (AttributeError, OSError):  # CPython and uPy have different errors
    print("Exception")

# dump to an object not allowed
try:
    json.dump(123, {})
except (AttributeError, OSError):  # CPython and uPy have different errors
    print("Exception")



# Test for VfsLittle using a RAM device, testing error handling

try:
    import vfs

    vfs.VfsLfs1
    vfs.VfsLfs2
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


class RAMBlockDevice:
    ERASE_BLOCK_SIZE = 1024

    def __init__(self, blocks):
        self.data = bytearray(blocks * self.ERASE_BLOCK_SIZE)

    def readblocks(self, block, buf, off):
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            buf[i] = self.data[addr + i]

    def writeblocks(self, block, buf, off):
        addr = block * self.ERASE_BLOCK_SIZE + off
        for i in range(len(buf)):
            self.data[addr + i] = buf[i]

    def ioctl(self, op, arg):
        if op == 4:  # block count
            return len(self.data) // self.ERASE_BLOCK_SIZE
        if op == 5:  # block size
            return self.ERASE_BLOCK_SIZE
        if op == 6:  # erase block
            return 0


def test(bdev, vfs_class):
    print("test", vfs_class)

    # mkfs with too-small block device
    try:
        vfs_class.mkfs(RAMBlockDevice(1))
    except OSError:
        print("mkfs OSError")

    # mount with invalid filesystem
    try:
        vfs_class(bdev)
    except OSError:
        print("mount OSError")

    # set up for following tests
    vfs_class.mkfs(bdev)
    fs = vfs_class(bdev)
    with fs.open("testfile", "w") as f:
        f.write("test")
    fs.mkdir("testdir")

    # ilistdir
    try:
        fs.ilistdir("noexist")
    except OSError:
        print("ilistdir OSError")

    # remove
    try:
        fs.remove("noexist")
    except OSError:
        print("remove OSError")

    # rmdir
    try:
        fs.rmdir("noexist")
    except OSError:
        print("rmdir OSError")

    # rename
    try:
        fs.rename("noexist", "somethingelse")
    except OSError:
        print("rename OSError")

    # mkdir
    try:
        fs.mkdir("testdir")
    except OSError:
        print("mkdir OSError")

    # chdir to nonexistent
    try:
        fs.chdir("noexist")
    except OSError:
        print("chdir OSError")
    print(fs.getcwd())  # check still at root

    # chdir to file
    try:
        fs.chdir("testfile")
    except OSError:
        print("chdir OSError")
    print(fs.getcwd())  # check still at root

    # stat
    try:
        fs.stat("noexist")
    except OSError:
        print("stat OSError")

    # error during seek
    with fs.open("testfile", "r") as f:
        f.seek(1 << 30)  # SEEK_SET
        try:
            f.seek(1 << 30, 1)  # SEEK_CUR
        except OSError:
            print("seek OSError")


bdev = RAMBlockDevice(30)
test(bdev, vfs.VfsLfs1)
test(bdev, vfs.VfsLfs2)



try:
    import uasyncio
except ImportError:
    print("SKIP")
    raise SystemExit


# Sample of public symbols we expect to see from `asyncio`. Verify they're all
# available on `uasyncio`.
expected = [
    "CancelledError",
    "create_task",
    "current_task",
    "Event",
    "gather",
    "get_event_loop",
    "Lock",
    "Loop",
    "open_connection",
    "run",
    "run_until_complete",
    "sleep",
    "sleep_ms",
    "start_server",
    "StreamReader",
    "StreamWriter",
    "Task",
    "ThreadSafeFlag",
    "wait_for",
]

for e in expected:
    getattr(uasyncio, e)



# DS18x20 temperature sensor driver for MicroPython.
# MIT license; Copyright (c) 2016 Damien P. George

from micropython import const

_CONVERT = const(0x44)
_RD_SCRATCH = const(0xBE)
_WR_SCRATCH = const(0x4E)


class DS18X20:
    def __init__(self, onewire):
        self.ow = onewire
        self.buf = bytearray(9)

    def scan(self):
        return [rom for rom in self.ow.scan() if rom[0] in (0x10, 0x22, 0x28)]

    def convert_temp(self):
        self.ow.reset(True)
        self.ow.writebyte(self.ow.SKIP_ROM)
        self.ow.writebyte(_CONVERT)

    def read_scratch(self, rom):
        self.ow.reset(True)
        self.ow.select_rom(rom)
        self.ow.writebyte(_RD_SCRATCH)
        self.ow.readinto(self.buf)
        if self.ow.crc8(self.buf):
            raise Exception("CRC error")
        return self.buf

    def write_scratch(self, rom, buf):
        self.ow.reset(True)
        self.ow.select_rom(rom)
        self.ow.writebyte(_WR_SCRATCH)
        self.ow.write(buf)

    def read_temp(self, rom):
        buf = self.read_scratch(rom)
        if rom[0] == 0x10:
            if buf[1]:
                t = buf[0] >> 1 | 0x80
                t = -((~t + 1) & 0xFF)
            else:
                t = buf[0] >> 1
            return t - 0.25 + (buf[7] - buf[6]) / buf[7]
        else:
            t = buf[1] << 8 | buf[0]
            if t & 0x8000:  # sign bit set
                t = -((t ^ 0xFFFF) + 1)
            return t / 16



# Check that yield-from can work without heap allocation

import micropython


# Yielding from a function generator
def sub_gen(a):
    for i in range(a):
        yield i


def gen(g):
    yield from g


g = gen(sub_gen(4))
micropython.heap_lock()
print(next(g))
print(next(g))
micropython.heap_unlock()


# Yielding from a user iterator
class G:
    def __init__(self):
        self.value = 0

    def __iter__(self):
        return self

    def __next__(self):
        v = self.value
        self.value += 1
        return v


g = gen(G())
micropython.heap_lock()
print(next(g))
print(next(g))
micropython.heap_unlock()



# This test ensures that the scheduler doesn't trigger any assertions
# while dealing with concurrent access from multiple threads.

import _thread
import time
import micropython
import gc

try:
    micropython.schedule
except AttributeError:
    print("SKIP")
    raise SystemExit

gc.disable()

_NUM_TASKS = 10000
_TIMEOUT_MS = 10000

n = 0  # How many times the task successfully ran.
t = None  # Start time of test, assigned here to preallocate entry in globals dict.
thread_run = True  # If the thread should continue running.


def task(x):
    global n
    n += 1


def thread():
    while thread_run:
        try:
            micropython.schedule(task, None)
        except RuntimeError:
            # Queue full, back off.
            time.sleep_ms(10)


for i in range(8):
    try:
        _thread.start_new_thread(thread, ())
    except OSError:
        # System cannot create a new thead, so stop trying to create them.
        break

# Wait up to 10 seconds for 10000 tasks to be scheduled.
t = time.ticks_ms()
while n < _NUM_TASKS and time.ticks_diff(time.ticks_ms(), t) < _TIMEOUT_MS:
    pass

# Stop all threads.
thread_run = False
time.sleep_ms(20)

gc.enable()

if n < _NUM_TASKS:
    # Not all the tasks were scheduled, likely the scheduler stopped working.
    print(n)
else:
    print("PASS")



"""
This test need a set of pins which can be set as inputs and have no external
pull up or pull down connected.
GP12 and GP17 must be connected together
"""
from machine import Pin
import os

mch = os.uname().machine
if "LaunchPad" in mch:
    pin_map = [
        "GP24",
        "GP12",
        "GP14",
        "GP15",
        "GP16",
        "GP17",
        "GP28",
        "GP8",
        "GP6",
        "GP30",
        "GP31",
        "GP3",
        "GP0",
        "GP4",
        "GP5",
    ]
    max_af_idx = 15
elif "WiPy" in mch:
    pin_map = [
        "GP23",
        "GP24",
        "GP12",
        "GP13",
        "GP14",
        "GP9",
        "GP17",
        "GP28",
        "GP22",
        "GP8",
        "GP30",
        "GP31",
        "GP0",
        "GP4",
        "GP5",
    ]
    max_af_idx = 15
else:
    raise Exception("Board not supported!")

# test initial value
p = Pin("GP12", Pin.IN)
Pin("GP17", Pin.OUT, value=1)
print(p() == 1)
Pin("GP17", Pin.OUT, value=0)
print(p() == 0)


def test_noinit():
    for p in pin_map:
        pin = Pin(p)
        pin.value()


def test_pin_read(pull):
    # enable the pull resistor on all pins, then read the value
    for p in pin_map:
        pin = Pin(p, mode=Pin.IN, pull=pull)
    for p in pin_map:
        print(pin())


def test_pin_af():
    for p in pin_map:
        for af in Pin(p).alt_list():
            if af[1] <= max_af_idx:
                Pin(p, mode=Pin.ALT, alt=af[1])
                Pin(p, mode=Pin.ALT_OPEN_DRAIN, alt=af[1])


# test un-initialized pins
test_noinit()
# test with pull-up and pull-down
test_pin_read(Pin.PULL_UP)
test_pin_read(Pin.PULL_DOWN)

# test all constructor combinations
pin = Pin(pin_map[0])
pin = Pin(pin_map[0], mode=Pin.IN)
pin = Pin(pin_map[0], mode=Pin.OUT)
pin = Pin(pin_map[0], mode=Pin.IN, pull=Pin.PULL_DOWN)
pin = Pin(pin_map[0], mode=Pin.IN, pull=Pin.PULL_UP)
pin = Pin(pin_map[0], mode=Pin.OPEN_DRAIN, pull=Pin.PULL_UP)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_DOWN)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=None)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_UP)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.LOW_POWER)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.MED_POWER)
pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.HIGH_POWER)
pin = Pin(pin_map[0], mode=Pin.OUT, drive=pin.LOW_POWER)
pin = Pin(pin_map[0], Pin.OUT, Pin.PULL_DOWN)
pin = Pin(pin_map[0], Pin.ALT, Pin.PULL_UP)
pin = Pin(pin_map[0], Pin.ALT_OPEN_DRAIN, Pin.PULL_UP)
test_pin_af()  # try the entire af range on all pins

# test pin init and printing
pin = Pin(pin_map[0])
pin.init(mode=Pin.IN)
print(pin)
pin.init(Pin.IN, Pin.PULL_DOWN)
print(pin)
pin.init(mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.LOW_POWER)
print(pin)
pin.init(mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.HIGH_POWER)
print(pin)

# test value in OUT mode
pin = Pin(pin_map[0], mode=Pin.OUT)
pin.value(0)
pin.toggle()  # test toggle
print(pin())
pin.toggle()  # test toggle again
print(pin())
# test different value settings
pin(1)
print(pin.value())
pin(0)
print(pin.value())
pin.value(1)
print(pin())
pin.value(0)
print(pin())

# test all getters and setters
pin = Pin(pin_map[0], mode=Pin.OUT)
# mode
print(pin.mode() == Pin.OUT)
pin.mode(Pin.IN)
print(pin.mode() == Pin.IN)
# pull
pin.pull(None)
print(pin.pull() == None)
pin.pull(Pin.PULL_DOWN)
print(pin.pull() == Pin.PULL_DOWN)
# drive
pin.drive(Pin.MED_POWER)
print(pin.drive() == Pin.MED_POWER)
pin.drive(Pin.HIGH_POWER)
print(pin.drive() == Pin.HIGH_POWER)
# id
print(pin.id() == pin_map[0])

# all the next ones MUST raise
try:
    pin = Pin(pin_map[0], mode=Pin.OUT, pull=Pin.PULL_UP, drive=pin.IN)  # incorrect drive value
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], mode=Pin.LOW_POWER, pull=Pin.PULL_UP)  # incorrect mode value
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], mode=Pin.IN, pull=Pin.HIGH_POWER)  # incorrect pull value
except Exception:
    print("Exception")

try:
    pin = Pin("A0", Pin.OUT, Pin.PULL_DOWN)  # incorrect pin id
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], Pin.IN, Pin.PULL_UP, alt=0)  # af specified in GPIO mode
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], Pin.OUT, Pin.PULL_UP, alt=7)  # af specified in GPIO mode
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], Pin.ALT, Pin.PULL_UP, alt=0)  # incorrect af
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], Pin.ALT_OPEN_DRAIN, Pin.PULL_UP, alt=-1)  # incorrect af
except Exception:
    print("Exception")

try:
    pin = Pin(pin_map[0], Pin.ALT_OPEN_DRAIN, Pin.PULL_UP, alt=16)  # incorrect af
except Exception:
    print("Exception")

try:
    pin.mode(Pin.PULL_UP)  # incorrect pin mode
except Exception:
    print("Exception")

try:
    pin.pull(Pin.OUT)  # incorrect pull
except Exception:
    print("Exception")

try:
    pin.drive(Pin.IN)  # incorrect drive strength
except Exception:
    print("Exception")

try:
    pin.id("ABC")  # id cannot be set
except Exception:
    print("Exception")



# MicroPython LoRa modem driver time on air tests
# MIT license; Copyright (c) 2023 Angus Gratton
#
# LoRa is a registered trademark or service mark of Semtech Corporation or its affiliates.
#
# ## What is this?
#
# Host tests for the BaseModem.get_time_on_air_us() function. Theses against
# dummy test values produced by the Semtech "SX1261 LoRa Calculator" software,
# as downloaded from
# https://lora-developers.semtech.com/documentation/product-documents/
#
# The app notes for SX1276 (AN1200.3) suggest a similar calculator exists for that
# modem, but it doesn't appear to be available for download any more. I couldn't find
# an accurate calculator for SX1276, so manually calculated the SF5 & SF6 test cases below
# (other values should be the same as SX1262).
#
# ## Instructions
#
# These tests are intended to be run on a host PC via micropython unix port:
#
# cd /path/to/micropython-lib/micropython/lora
# micropython -m tests.test_time_on_air
#
# Note: Using the working directory shown above is easiest way to ensure 'lora' files are imported.
#
from lora import SX1262, SX1276

# Allow time calculations to deviate by up to this much as a ratio
# of the expected value (due to floating point, etc.)
TIME_ERROR_RATIO = 0.00001  # 0.001%


def main():
    sx1262 = SX1262(spi=DummySPI(), cs=DummyPin(), busy=DummyPin())
    sx1276 = SX1276(spi=DummySPI(0x12), cs=DummyPin())

    # Test case format is based on the layout of the Semtech Calculator UI:
    #
    # (modem_instance,
    #  (modem settings),
    #  [
    #      ((packet config), (output values)),
    #      ...
    #  ],
    # ),
    #
    # where each set of modem settings maps to zero or more packet config / output pairs
    #
    # - modem instance is sx1262 or sx1276 (SF5 & SF6 are different between these modems)
    # - (modem settings) is (sf, bw (in khz), coding_rate, low_datarate_optimize)
    # - (packet config) is (preamble_len, payload_len, explicit_header, crc_en)
    # - (output values) is (total_symbols_excl, symbol_time in ms, time_on_air in ms)
    #
    # NOTE: total_symbols_excl is the value shown in the calculator output,
    # which doesn't include 8 symbols of constant overhead between preamble and
    # header+payload+crc. I think this is a bug in the Semtech calculator(!).
    # These 8 symbols are included when the calculator derives the total time on
    # air.
    #
    # NOTE ALSO: The "symbol_time" only depends on the modem settings so is
    # repeated each group of test cases, and the "time_on_air" is the previous
    # two output values multiplied (after accounting for the 8 symbols noted
    # above). This repetition is deliberate to make the cases easier to read
    # line-by-line when comparing to the calculator window.
    CASES = [
        (
            sx1262,
            (12, 500, 5, False),  # Calculator defaults when launching calculator
            [
                ((8, 1, True, True), (17.25, 8.192, 206.848)),  # Calculator defaults
                ((12, 64, True, True), (71.25, 8.192, 649.216)),
                ((8, 1, True, False), (12.25, 8.192, 165.888)),
                ((8, 192, True, True), (172.25, 8.192, 1476.608)),
                ((12, 16, False, False), (26.25, 8.192, 280.576)),
            ],
        ),
        (
            sx1262,
            (8, 125, 6, False),
            [
                ((8, 1, True, True), (18.25, 2.048, 53.760)),
                ((8, 2, True, True), (18.25, 2.048, 53.760)),
                ((8, 2, True, False), (18.25, 2.048, 53.760)),
                ((8, 3, True, True), (24.25, 2.048, 66.048)),
                ((8, 3, True, False), (18.25, 2.048, 53.760)),
                ((8, 4, True, True), (24.25, 2.048, 66.048)),
                ((8, 4, True, False), (18.25, 2.048, 53.760)),
                ((8, 5, True, True), (24.25, 2.048, 66.048)),
                ((8, 5, True, False), (24.25, 2.048, 66.048)),
                ((8, 253, True, True), (396.25, 2.048, 827.904)),
                ((8, 253, True, False), (396.25, 2.048, 827.904)),
                ((12, 5, False, True), (22.25, 2.048, 61.952)),
                ((12, 5, False, False), (22.25, 2.048, 61.952)),
                ((12, 10, False, True), (34.25, 2.048, 86.528)),
                ((12, 253, False, True), (394.25, 2.048, 823.808)),
            ],
        ),
        # quick check that sx1276 is the same as sx1262 for SF>6
        (
            sx1276,
            (8, 125, 6, False),
            [
                ((8, 1, True, True), (18.25, 2.048, 53.760)),
                ((8, 2, True, True), (18.25, 2.048, 53.760)),
                ((12, 5, False, True), (22.25, 2.048, 61.952)),
                ((12, 5, False, False), (22.25, 2.048, 61.952)),
            ],
        ),
        # SF5 on SX1262
        (
            sx1262,
            (5, 500, 5, False),
            [
                (
                    (2, 1, True, False),
                    (13.25, 0.064, 1.360),
                ),  # Shortest possible LoRa packet?
                ((2, 1, True, True), (18.25, 0.064, 1.680)),
                ((12, 1, False, False), (18.25, 0.064, 1.680)),
                ((12, 253, False, True), (523.25, 0.064, 34.000)),
            ],
        ),
        (
            sx1262,
            (5, 125, 8, False),
            [
                ((12, 253, False, True), (826.25, 0.256, 213.568)),
            ],
        ),
        # SF5 on SX1276
        #
        # Note: SF5 & SF6 settings are different between SX1262 & SX1276.
        #
        # There's no Semtech official calculator available for SX1276, so the
        # symbol length is calculated by copying the formula from the datasheet
        # "Time on air" section. Symbol time is the same as SX1262. Then the
        # time on air is manually calculated by multiplying the two together.
        #
        # see the functions sx1276_num_payload and sx1276_num_symbols at end of this module
        # for the actual functions used.
        (
            sx1276,
            (5, 500, 5, False),
            [
                (
                    (2, 1, True, False),
                    (19.25 - 8, 0.064, 1.232),
                ),  # Shortest possible LoRa packet?
                ((2, 1, True, True), (24.25 - 8, 0.064, 1.552)),
                ((12, 1, False, False), (24.25 - 8, 0.064, 1.552)),
                ((12, 253, False, True), (534.25 - 8, 0.064, 34.192)),
            ],
        ),
        (
            sx1276,
            (5, 125, 8, False),
            [
                ((12, 253, False, True), (840.25 - 8, 0.256, 215.104)),
            ],
        ),
        (
            sx1262,
            (12, 7.81, 8, True),  # Slowest possible
            [
                ((128, 253, True, True), (540.25, 524.456, 287532.907)),
                ((1000, 253, True, True), (1412.25, 524.456, 744858.387)),
            ],
        ),
        (
            sx1262,
            (11, 10.42, 7, True),
            [
                ((25, 16, True, True), (57.25, 196.545, 12824.568)),
                ((25, 16, False, False), (50.25, 196.545, 11448.752)),
            ],
        ),
    ]

    tests = 0
    failures = set()
    for modem, modem_settings, packets in CASES:
        (sf, bw_khz, coding_rate, low_datarate_optimize) = modem_settings
        print(
            f"Modem config sf={sf} bw={bw_khz}kHz coding_rate=4/{coding_rate} "
            + f"low_datarate_optimize={low_datarate_optimize}"
        )

        # We don't call configure() as the Dummy interfaces won't handle it,
        # just update the BaseModem fields directly
        modem._sf = sf
        modem._bw_hz = int(bw_khz * 1000)
        modem._coding_rate = coding_rate

        # Low datarate optimize on/off is auto-configured in the current driver,
        # check the automatic selection matches the test case from the
        # calculator
        if modem._get_ldr_en() != low_datarate_optimize:
            print(
                f" -- ERROR: Test case has low_datarate_optimize={low_datarate_optimize} "
                + f"but modem selects {modem._get_ldr_en()}"
            )
            failures += 1
            continue  # results will not match so don't run any of the packet test cases

        for packet_config, expected_outputs in packets:
            preamble_len, payload_len, explicit_header, crc_en = packet_config
            print(
                f" -- preamble_len={preamble_len} payload_len={payload_len} "
                + f"explicit_header={explicit_header} crc_en={crc_en}"
            )
            modem._preamble_len = preamble_len
            modem._implicit_header = not explicit_header  # opposite logic to calculator
            modem._crc_en = crc_en

            # Now calculate the symbol length and times and compare with the expected valuesd
            (
                expected_symbols,
                expected_symbol_time,
                expected_time_on_air,
            ) = expected_outputs

            print(f" ---- calculator shows total length {expected_symbols}")
            expected_symbols += 8  # Account for the calculator bug mentioned in the comment above

            n_symbols = modem.get_n_symbols_x4(payload_len) / 4.0
            symbol_time_us = modem._get_t_sym_us()
            time_on_air_us = modem.get_time_on_air_us(payload_len)

            tests += 1

            if n_symbols == expected_symbols:
                print(f" ---- symbols {n_symbols}")
            else:
                print(f" ---- SYMBOL COUNT ERROR expected {expected_symbols} got {n_symbols}")
                failures.add((modem, modem_settings, packet_config))

            max_error = expected_symbol_time * 1000 * TIME_ERROR_RATIO
            if abs(int(expected_symbol_time * 1000) - symbol_time_us) <= max_error:
                print(f" ---- symbol time {expected_symbol_time}ms")
            else:
                print(
                    f" ---- SYMBOL TIME ERROR expected {expected_symbol_time}ms "
                    + f"got {symbol_time_us}us"
                )
                failures.add((modem, modem_settings, packet_config))

            max_error = expected_time_on_air * 1000 * TIME_ERROR_RATIO
            if abs(int(expected_time_on_air * 1000) - time_on_air_us) <= max_error:
                print(f" ---- time on air {expected_time_on_air}ms")
            else:
                print(
                    f" ---- TIME ON AIR ERROR expected {expected_time_on_air}ms "
                    + f"got {time_on_air_us}us"
                )
                failures.add((modem, modem_settings, packet_config))

        print("************************")

    print(f"\n{len(failures)}/{tests} tests failed")
    if failures:
        print("FAILURES:")
        for f in failures:
            print(f)
        raise SystemExit(1)
    print("SUCCESS")


class DummySPI:
    # Dummy SPI Interface allows us to use normal constructors
    #
    # Reading will always return the 'always_read' value
    def __init__(self, always_read=0x00):
        self.always_read = always_read

    def write_readinto(self, _wrbuf, rdbuf):
        for i in range(len(rdbuf)):
            rdbuf[i] = self.always_read


class DummyPin:
    # Dummy Pin interface allows us to use normal constructors
    def __init__(self):
        pass

    def __call__(self, _=None):
        pass


# Copies of the functions used to calculate SX1276 SF5, SF6 test case symbol counts.
# (see comments above).
#
# These are written as closely to the SX1276 datasheet "Time on air" section as
# possible, quite different from the BaseModem implementation.


def sx1276_n_payload(pl, sf, ih, de, cr, crc):
    import math

    ceil_arg = 8 * pl - 4 * sf + 28 + 16 * crc - 20 * ih
    ceil_arg /= 4 * (sf - 2 * de)
    return 8 + max(math.ceil(ceil_arg) * (cr + 4), 0)


def sx1276_n_syms(pl, sf, ih, de, cr, crc, n_preamble):
    return sx1276_n_payload(pl, sf, ih, de, cr, crc) + n_preamble + 4.25


if __name__ == "__main__":
    main()



# test calling a function with *tuple and **dict

def f(a, b, c, d):
    print(a, b, c, d)

f(*(1, 2), **{'c':3, 'd':4})
f(*(1, 2), **{['c', 'd'][i]:(3 + i) for i in range(2)})

try:
    eval("f(**{'a': 1}, *(2, 3, 4))")
except SyntaxError:
    print("SyntaxError")

# test calling a method with *tuple and **dict

class A:
    def f(self, a, b, c, d):
        print(a, b, c, d)

a = A()
a.f(*(1, 2), **{'c':3, 'd':4})
a.f(*(1, 2), **{['c', 'd'][i]:(3 + i) for i in range(2)})

try:
    eval("a.f(**{'a': 1}, *(2, 3, 4))")
except SyntaxError:
    print("SyntaxError")


# coverage test for arg allocation corner case

def f2(*args, **kwargs):
    print(len(args), len(kwargs))


f2(*iter(range(4)), **{'a': 1})

# case where *args is not a tuple/list and takes up most of the memory allocated for **kwargs
f2(*iter(range(100)), **{str(i): i for i in range(100)})

# regression test - iterable with unknown len() was exactly using preallocated
# memory causing args 4 and 5 to overflow the allocated arg array
print(1, *iter((1, 2, 3)), *iter((1, 2, 3)), 4, 5, sep=",")



# The asyncio package is built from the standard implementation but with the
# core scheduler replaced with a custom scheduler that uses the JavaScript
# runtime (with setTimeout an Promise's) to contrtol the scheduling.

package(
    "asyncio",
    (
        "event.py",
        "funcs.py",
        "lock.py",
    ),
    base_path="$(MPY_DIR)/extmod",
    opt=3,
)

package(
    "asyncio",
    (
        "__init__.py",
        "core.py",
    ),
    base_path="$(PORT_DIR)",
    opt=3,
)



# test that socket.connect() on a non-blocking socket raises EINPROGRESS
# and that an immediate write/send/read/recv does the right thing

import sys, time, socket, errno, ssl

isMP = sys.implementation.name == "micropython"


def dp(e):
    # uncomment next line for development and testing, to print the actual exceptions
    # print(repr(e))
    pass


# do_connect establishes the socket and wraps it if tls is True.
# If handshake is true, the initial connect (and TLS handshake) is
# allowed to be performed before returning.
def do_connect(peer_addr, tls, handshake):
    s = socket.socket()
    s.setblocking(False)
    try:
        # print("Connecting to", peer_addr)
        s.connect(peer_addr)
    except OSError as er:
        print("connect:", er.errno == errno.EINPROGRESS)
        if er.errno != errno.EINPROGRESS:
            print("  got", er.errno)
    # wrap with ssl/tls if desired
    if tls:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        if hasattr(ssl_context, "check_hostname"):
            ssl_context.check_hostname = False

        try:
            s = ssl_context.wrap_socket(s, do_handshake_on_connect=handshake)
            print("wrap: True")
        except Exception as e:
            dp(e)
            print("wrap:", e)
    elif handshake:
        # just sleep a little bit, this allows any connect() errors to happen
        time.sleep(0.2)
    return s


# test runs the test against a specific peer address.
def test(peer_addr, tls=False, handshake=False):
    # MicroPython plain sockets have read/write, but CPython's don't
    # MicroPython TLS sockets and CPython's have read/write
    # hasRW captures this wonderful state of affairs
    hasRW = isMP or tls

    # MicroPython plain sockets and CPython's have send/recv
    # MicroPython TLS sockets don't have send/recv, but CPython's do
    # hasSR captures this wonderful state of affairs
    hasSR = not (isMP and tls)

    # connect + send
    if hasSR:
        s = do_connect(peer_addr, tls, handshake)
        # send -> 4 or EAGAIN
        try:
            ret = s.send(b"1234")
            print("send:", handshake and ret == 4)
        except OSError as er:
            #
            dp(er)
            print("send:", er.errno in (errno.EAGAIN, errno.EINPROGRESS))
        s.close()
    else:  # fake it...
        print("connect:", True)
        if tls:
            print("wrap:", True)
        print("send:", True)

    # connect + write
    if hasRW:
        s = do_connect(peer_addr, tls, handshake)
        # write -> None
        try:
            ret = s.write(b"1234")
            print("write:", ret in (4, None))  # SSL may accept 4 into buffer
        except OSError as er:
            dp(er)
            print("write:", False)  # should not raise
        except ValueError as er:  # CPython
            dp(er)
            print("write:", er.args[0] == "Write on closed or unwrapped SSL socket.")
        s.close()
    else:  # fake it...
        print("connect:", True)
        if tls:
            print("wrap:", True)
        print("write:", True)

    if hasSR:
        # connect + recv
        s = do_connect(peer_addr, tls, handshake)
        # recv -> EAGAIN
        try:
            print("recv:", s.recv(10))
        except OSError as er:
            dp(er)
            print("recv:", er.errno == errno.EAGAIN)
        s.close()
    else:  # fake it...
        print("connect:", True)
        if tls:
            print("wrap:", True)
        print("recv:", True)

    # connect + read
    if hasRW:
        s = do_connect(peer_addr, tls, handshake)
        # read -> None
        try:
            ret = s.read(10)
            print("read:", ret is None)
        except OSError as er:
            dp(er)
            print("read:", False)  # should not raise
        except ValueError as er:  # CPython
            dp(er)
            print("read:", er.args[0] == "Read on closed or unwrapped SSL socket.")
        s.close()
    else:  # fake it...
        print("connect:", True)
        if tls:
            print("wrap:", True)
        print("read:", True)


if __name__ == "__main__":
    # these tests use a non-existent test IP address, this way the connect takes forever and
    # we can see EAGAIN/None (https://tools.ietf.org/html/rfc5737)
    print("--- Plain sockets to nowhere ---")
    test(socket.getaddrinfo("192.0.2.1", 80)[0][-1], False, False)
    print("--- SSL sockets to nowhere ---")
    # this test fails with AXTLS because do_handshake=False blocks on first read/write and
    # there it times out until the connect is aborted
    test(socket.getaddrinfo("192.0.2.1", 443)[0][-1], True, False)
    print("--- Plain sockets ---")
    test(socket.getaddrinfo("micropython.org", 80)[0][-1], False, True)
    print("--- SSL sockets ---")
    test(socket.getaddrinfo("micropython.org", 443)[0][-1], True, True)



#!/usr/bin/env python

"""
The WiPy firmware update script. Transmits the specified firmware file
over FTP, and then resets the WiPy and optionally verifies that software
was correctly updated.

Usage:

    ./update-wipy.py --file "path_to_mcuimg.bin" --verify

Or:

    python update-wipy.py --file "path_to_mcuimg.bin"
"""

import sys
import argparse
import time
import socket
from ftplib import FTP
from telnetlib import Telnet


def print_exception(e):
    print("Exception: {}, on line {}".format(e, sys.exc_info()[-1].tb_lineno))


def ftp_directory_exists(ftpobj, directory_name):
    filelist = []
    ftpobj.retrlines("LIST", filelist.append)
    for f in filelist:
        if f.split()[-1] == directory_name:
            return True
    return False


def transfer_file(args):
    with FTP(args.ip, timeout=20) as ftp:
        print("FTP connection established")

        if "230" in ftp.login(args.user, args.password):
            print("Login successful")

            if "250" in ftp.cwd("/flash"):
                if not ftp_directory_exists(ftp, "sys"):
                    print("/flash/sys directory does not exist")
                    if "550" not in ftp.mkd("sys"):
                        print("/flash/sys directory created")
                    else:
                        print("Error: cannot create /flash/sys directory")
                        return False
                if "250" in ftp.cwd("sys"):
                    print("Entered '/flash/sys' directory")
                    with open(args.file, "rb") as fwfile:
                        print("Firmware image found, initiating transfer...")
                        if "226" in ftp.storbinary("STOR " + "mcuimg.bin", fwfile, 512):
                            print("File transfer complete")
                            return True
                        else:
                            print("Error: file transfer failed")
                else:
                    print("Error: cannot enter /flash/sys directory")
            else:
                print("Error: cannot enter /flash directory")
        else:
            print("Error: ftp login failed")

    return False


def reset_board(args):
    success = False

    try:
        tn = Telnet(args.ip, timeout=5)
        print("Connected via Telnet, trying to login now")

        if b"Login as:" in tn.read_until(b"Login as:", timeout=5):
            tn.write(bytes(args.user, "ascii") + b"\r\n")

            if b"Password:" in tn.read_until(b"Password:", timeout=5):
                # needed because of internal implementation details of the WiPy's telnet server
                time.sleep(0.2)
                tn.write(bytes(args.password, "ascii") + b"\r\n")

                if b'Type "help()" for more information.' in tn.read_until(
                    b'Type "help()" for more information.', timeout=5
                ):
                    print("Telnet login succeeded")
                    tn.write(b"\r\x03\x03")  # ctrl-C twice: interrupt any running program
                    time.sleep(1)
                    tn.write(b"\r\x02")  # ctrl-B: enter friendly REPL
                    if b'Type "help()" for more information.' in tn.read_until(
                        b'Type "help()" for more information.', timeout=5
                    ):
                        tn.write(b"import machine\r\n")
                        tn.write(b"machine.reset()\r\n")
                        time.sleep(2)
                        print("Reset performed")
                        success = True
                    else:
                        print("Error: cannot enter friendly REPL")
                else:
                    print("Error: telnet login failed")

    except Exception as e:
        print_exception(e)
    finally:
        try:
            tn.close()
        except Exception:
            pass
        return success


def verify_update(args):
    success = False
    firmware_tag = ""

    def find_tag(tag):
        if tag in firmware_tag:
            print("Verification passed")
            return True
        else:
            print("Error: verification failed, the git tag doesn't match")
            return False

    retries = 0
    while True:
        try:
            # Specify a longer time out value here because the board has just been
            # reset and the wireless connection might not be fully established yet
            tn = Telnet(args.ip, timeout=10)
            print("Connected via telnet again, lets check the git tag")
            break
        except socket.timeout:
            if retries < 5:
                print("Timeout while connecting via telnet, retrying...")
                retries += 1
            else:
                print("Error: Telnet connection timed out!")
                return False

    try:
        firmware_tag = tn.read_until(b"with CC3200")
        tag_file_path = args.file.rstrip("mcuimg.bin") + "genhdr/mpversion.h"

        if args.tag is not None:
            success = find_tag(bytes(args.tag, "ascii"))
        else:
            with open(tag_file_path) as tag_file:
                for line in tag_file:
                    bline = bytes(line, "ascii")
                    if b"MICROPY_GIT_TAG" in bline:
                        bline = (
                            bline.lstrip(b"#define MICROPY_GIT_TAG ")
                            .replace(b'"', b"")
                            .replace(b"\r", b"")
                            .replace(b"\n", b"")
                        )
                        success = find_tag(bline)
                        break

    except Exception as e:
        print_exception(e)
    finally:
        try:
            tn.close()
        except Exception:
            pass
        return success


def main():
    cmd_parser = argparse.ArgumentParser(
        description="Update the WiPy firmware with the specified image file"
    )
    cmd_parser.add_argument("-f", "--file", default=None, help="the path of the firmware file")
    cmd_parser.add_argument("-u", "--user", default="micro", help="the username")
    cmd_parser.add_argument("-p", "--password", default="python", help="the login password")
    cmd_parser.add_argument("--ip", default="192.168.1.1", help="the ip address of the WiPy")
    cmd_parser.add_argument(
        "--verify", action="store_true", help="verify that the update succeeded"
    )
    cmd_parser.add_argument("-t", "--tag", default=None, help="git tag of the firmware image")
    args = cmd_parser.parse_args()

    result = 1

    try:
        if args.file is None:
            raise ValueError("the image file path must be specified")
        if transfer_file(args):
            if reset_board(args):
                if args.verify:
                    print("Waiting for the WiFi connection to come up again...")
                    # this time is to allow the system's wireless network card to
                    # connect to the WiPy again.
                    time.sleep(5)
                    if verify_update(args):
                        result = 0
                else:
                    result = 0

    except Exception as e:
        print_exception(e)
    finally:
        sys.exit(result)


if __name__ == "__main__":
    main()



# MicroPython USB MIDI module
# MIT license; Copyright (c) 2023 Paul Hamshere, 2023-2024 Angus Gratton
from micropython import const, schedule
import struct

from .core import Interface, Buffer

_EP_IN_FLAG = const(1 << 7)

_INTERFACE_CLASS_AUDIO = const(0x01)
_INTERFACE_SUBCLASS_AUDIO_CONTROL = const(0x01)
_INTERFACE_SUBCLASS_AUDIO_MIDISTREAMING = const(0x03)

# Audio subclass extends the standard endpoint descriptor
# with two extra bytes
_STD_DESC_AUDIO_ENDPOINT_LEN = const(9)
_CLASS_DESC_ENDPOINT_LEN = const(5)

_STD_DESC_ENDPOINT_TYPE = const(0x5)

_JACK_TYPE_EMBEDDED = const(0x01)
_JACK_TYPE_EXTERNAL = const(0x02)

_JACK_IN_DESC_LEN = const(6)
_JACK_OUT_DESC_LEN = const(9)

# MIDI Status bytes. For Channel messages these are only the upper 4 bits, ORed with the channel number.
# As per https://www.midi.org/specifications-old/item/table-1-summary-of-midi-message
_MIDI_NOTE_OFF = const(0x80)
_MIDI_NOTE_ON = const(0x90)
_MIDI_POLY_KEYPRESS = const(0xA0)
_MIDI_CONTROL_CHANGE = const(0xB0)

# USB-MIDI CINs (Code Index Numbers), as per USB MIDI Table 4-1
_CIN_SYS_COMMON_2BYTE = const(0x2)
_CIN_SYS_COMMON_3BYTE = const(0x3)
_CIN_SYSEX_START = const(0x4)
_CIN_SYSEX_END_1BYTE = const(0x5)
_CIN_SYSEX_END_2BYTE = const(0x6)
_CIN_SYSEX_END_3BYTE = const(0x7)
_CIN_NOTE_OFF = const(0x8)
_CIN_NOTE_ON = const(0x9)
_CIN_POLY_KEYPRESS = const(0xA)
_CIN_CONTROL_CHANGE = const(0xB)
_CIN_PROGRAM_CHANGE = const(0xC)
_CIN_CHANNEL_PRESSURE = const(0xD)
_CIN_PITCH_BEND = const(0xE)
_CIN_SINGLE_BYTE = const(0xF)  # Not currently supported

# Jack IDs for a simple bidrectional MIDI device(!)
_EMB_IN_JACK_ID = const(1)
_EXT_IN_JACK_ID = const(2)
_EMB_OUT_JACK_ID = const(3)
_EXT_OUT_JACK_ID = const(4)

# Data flows, as modelled by USB-MIDI and this hypothetical interface, are as follows:
# Device RX = USB OUT EP => _EMB_IN_JACK => _EMB_OUT_JACK
# Device TX = _EXT_IN_JACK => _EMB_OUT_JACK => USB IN EP


class MIDIInterface(Interface):
    # Base class to implement a USB MIDI device in Python.
    #
    # To be compliant this also regisers a dummy USB Audio interface, but that
    # interface isn't otherwise used.

    def __init__(self, rxlen=16, txlen=16):
        # Arguments are size of transmit and receive buffers in bytes.

        super().__init__()
        self.ep_out = None  # Set during enumeration. RX direction (host to device)
        self.ep_in = None  # TX direction (device to host)
        self._rx = Buffer(rxlen)
        self._tx = Buffer(txlen)

    # Callbacks for handling received MIDI messages.
    #
    # Subclasses can choose between overriding on_midi_event
    # and handling all MIDI events manually, or overriding the
    # functions for note on/off and control change, only.

    def on_midi_event(self, cin, midi0, midi1, midi2):
        ch = midi0 & 0x0F
        if cin == _CIN_NOTE_ON:
            self.on_note_on(ch, midi1, midi2)
        elif cin == _CIN_NOTE_OFF:
            self.on_note_off(ch, midi1, midi2)
        elif cin == _CIN_CONTROL_CHANGE:
            self.on_control_change(ch, midi1, midi2)

    def on_note_on(self, channel, pitch, vel):
        pass  # Override to handle Note On messages

    def on_note_off(self, channel, pitch, vel):
        pass  # Override to handle Note On messages

    def on_control_change(self, channel, controller, value):
        pass  # Override to handle Control Change messages

    # Helper functions for sending common MIDI messages

    def note_on(self, channel, pitch, vel=0x40):
        self.send_event(_CIN_NOTE_ON, _MIDI_NOTE_ON | channel, pitch, vel)

    def note_off(self, channel, pitch, vel=0x40):
        self.send_event(_CIN_NOTE_OFF, _MIDI_NOTE_OFF | channel, pitch, vel)

    def control_change(self, channel, controller, value):
        self.send_event(_CIN_CONTROL_CHANGE, _MIDI_CONTROL_CHANGE | channel, controller, value)

    def send_event(self, cin, midi0, midi1=0, midi2=0):
        # Queue a MIDI Event Packet to send to the host.
        #
        # CIN = USB-MIDI Code Index Number, see USB MIDI 1.0 section 4 "USB-MIDI Event Packets"
        #
        # Remaining arguments are 0-3 MIDI data bytes.
        #
        # Note this function returns when the MIDI Event Packet has been queued,
        # not when it's been received by the host.
        #
        # Returns False if the TX buffer is full and the MIDI Event could not be queued.
        w = self._tx.pend_write()
        if len(w) < 4:
            return False  # TX buffer is full. TODO: block here?
        w[0] = cin  # leave cable number as 0?
        w[1] = midi0
        w[2] = midi1
        w[3] = midi2
        self._tx.finish_write(4)
        self._tx_xfer()
        return True

    def _tx_xfer(self):
        # Keep an active IN transfer to send data to the host, whenever
        # there is data to send.
        if self.is_open() and not self.xfer_pending(self.ep_in) and self._tx.readable():
            self.submit_xfer(self.ep_in, self._tx.pend_read(), self._tx_cb)

    def _tx_cb(self, ep, res, num_bytes):
        if res == 0:
            self._tx.finish_read(num_bytes)
        self._tx_xfer()

    def _rx_xfer(self):
        # Keep an active OUT transfer to receive MIDI events from the host
        if self.is_open() and not self.xfer_pending(self.ep_out) and self._rx.writable():
            self.submit_xfer(self.ep_out, self._rx.pend_write(), self._rx_cb)

    def _rx_cb(self, ep, res, num_bytes):
        if res == 0:
            self._rx.finish_write(num_bytes)
            schedule(self._on_rx, None)
        self._rx_xfer()

    def on_open(self):
        super().on_open()
        # kick off any transfers that may have queued while the device was not open
        self._tx_xfer()
        self._rx_xfer()

    def _on_rx(self, _):
        # Receive MIDI events. Called via micropython.schedule, outside of the USB callback function.
        m = self._rx.pend_read()
        i = 0
        while i <= len(m) - 4:
            cin = m[i] & 0x0F
            self.on_midi_event(cin, m[i + 1], m[i + 2], m[i + 3])
            i += 4
        self._rx.finish_read(i)

    def desc_cfg(self, desc, itf_num, ep_num, strs):
        # Start by registering a USB Audio Control interface, that is required to point to the
        # actual MIDI interface
        desc.interface(itf_num, 0, _INTERFACE_CLASS_AUDIO, _INTERFACE_SUBCLASS_AUDIO_CONTROL)

        # Append the class-specific AudioControl interface descriptor
        desc.pack(
            "<BBBHHBB",
            9,  # bLength
            0x24,  # bDescriptorType CS_INTERFACE
            0x01,  # bDescriptorSubtype MS_HEADER
            0x0100,  # BcdADC
            0x0009,  # wTotalLength
            0x01,  # bInCollection,
            itf_num + 1,  # baInterfaceNr - points to the MIDI Streaming interface
        )

        # Next add the MIDI Streaming interface descriptor
        desc.interface(
            itf_num + 1, 2, _INTERFACE_CLASS_AUDIO, _INTERFACE_SUBCLASS_AUDIO_MIDISTREAMING
        )

        # Append the class-specific interface descriptors

        # Midi Streaming interface descriptor
        desc.pack(
            "<BBBHH",
            7,  # bLength
            0x24,  # bDescriptorType CS_INTERFACE
            0x01,  # bDescriptorSubtype MS_HEADER
            0x0100,  # BcdADC
            # wTotalLength: of all class-specific descriptors
            7
            + 2
            * (
                _JACK_IN_DESC_LEN
                + _JACK_OUT_DESC_LEN
                + _STD_DESC_AUDIO_ENDPOINT_LEN
                + _CLASS_DESC_ENDPOINT_LEN
            ),
        )

        # The USB MIDI standard 1.0 allows modelling a baffling range of MIDI
        # devices with different permutations of Jack descriptors, with a lot of
        # scope for indicating internal connections in the device (as
        # "virtualised" by the USB MIDI standard). Much of the options don't
        # really change the USB behaviour but provide metadata to the host.
        #
        # As observed elsewhere online, the standard ends up being pretty
        # complex and unclear in parts, but there is a clear simple example in
        # an Appendix. So nearly everyone implements the device from the
        # Appendix as-is, even when it's not a good fit for their application,
        # and ignores the rest of the standard.
        #
        # For now, this is what this class does as well.

        _jack_in_desc(desc, _JACK_TYPE_EMBEDDED, _EMB_IN_JACK_ID)
        _jack_in_desc(desc, _JACK_TYPE_EXTERNAL, _EXT_IN_JACK_ID)
        _jack_out_desc(desc, _JACK_TYPE_EMBEDDED, _EMB_OUT_JACK_ID, _EXT_IN_JACK_ID, 1)
        _jack_out_desc(desc, _JACK_TYPE_EXTERNAL, _EXT_OUT_JACK_ID, _EMB_IN_JACK_ID, 1)

        # One MIDI endpoint in each direction, plus the
        # associated CS descriptors

        self.ep_out = ep_num
        self.ep_in = ep_num | _EP_IN_FLAG

        # rx side, USB "in" endpoint and embedded MIDI IN Jacks
        _audio_endpoint(desc, self.ep_in, _EMB_OUT_JACK_ID)

        # tx side, USB "out" endpoint and embedded MIDI OUT jacks
        _audio_endpoint(desc, self.ep_out, _EMB_IN_JACK_ID)

    def num_itfs(self):
        return 2

    def num_eps(self):
        return 1


def _jack_in_desc(desc, bJackType, bJackID):
    # Utility function appends a "JACK IN" descriptor with
    # specified bJackType and bJackID
    desc.pack(
        "<BBBBBB",
        _JACK_IN_DESC_LEN,  # bLength
        0x24,  # bDescriptorType CS_INTERFACE
        0x02,  # bDescriptorSubtype MIDI_IN_JACK
        bJackType,
        bJackID,
        0x00,  # iJack, no string descriptor support yet
    )


def _jack_out_desc(desc, bJackType, bJackID, bSourceId, bSourcePin):
    # Utility function appends a "JACK IN" descriptor with
    # specified bJackType and bJackID
    desc.pack(
        "<BBBBBBBBB",
        _JACK_OUT_DESC_LEN,  # bLength
        0x24,  # bDescriptorType CS_INTERFACE
        0x03,  # bDescriptorSubtype MIDI_OUT_JACK
        bJackType,
        bJackID,
        0x01,  # bNrInputPins
        bSourceId,  # baSourceID(1)
        bSourcePin,  # baSourcePin(1)
        0x00,  # iJack, no string descriptor support yet
    )


def _audio_endpoint(desc, bEndpointAddress, emb_jack_id):
    # Append a standard USB endpoint descriptor and the USB class endpoint descriptor
    # for this endpoint.
    #
    # Audio Class devices extend the standard endpoint descriptor with two extra bytes,
    # so we can't easily call desc.endpoint() for the first part.
    desc.pack(
        # Standard USB endpoint descriptor (plus audio tweaks)
        "<BBBBHBBB"
        # Class endpoint descriptor
        "BBBBB",
        _STD_DESC_AUDIO_ENDPOINT_LEN,  # wLength
        _STD_DESC_ENDPOINT_TYPE,  # bDescriptorType
        bEndpointAddress,
        2,  # bmAttributes, bulk
        64,  # wMaxPacketSize
        0,  # bInterval
        0,  # bRefresh (unused)
        0,  # bSynchInterval (unused)
        _CLASS_DESC_ENDPOINT_LEN,  # bLength
        0x25,  # bDescriptorType CS_ENDPOINT
        0x01,  # bDescriptorSubtype MS_GENERAL
        1,  # bNumEmbMIDIJack
        emb_jack_id,  # BaAssocJackID(1)
    )



include("$(PORT_DIR)/boards/manifest.py")

# Networking
require("bundle-networking")

# Utils
require("time")
require("senml")
require("logging")

# Bluetooth
require("aioble")

# Register external library
add_library("arduino-lib", "$(ARDUINO_LIB_DIR)")

# RPC
require("msgpackrpc", library="arduino-lib")

# Lora/CMWX1 driver
require("cmwx1", library="arduino-lib")

# SE05x driver
require("se05x", library="arduino-lib")



try:
    raise ArithmeticError
except Exception:
    print("Caught ArithmeticError via Exception")

try:
    raise ArithmeticError
except ArithmeticError:
    print("Caught ArithmeticError")

try:
    raise AssertionError
except Exception:
    print("Caught AssertionError via Exception")

try:
    raise AssertionError
except AssertionError:
    print("Caught AssertionError")

try:
    raise AttributeError
except Exception:
    print("Caught AttributeError via Exception")

try:
    raise AttributeError
except AttributeError:
    print("Caught AttributeError")

try:
    raise EOFError
except Exception:
    print("Caught EOFError via Exception")

try:
    raise EOFError
except EOFError:
    print("Caught EOFError")

try:
    raise Exception
except BaseException:
    print("Caught Exception via BaseException")

try:
    raise Exception
except Exception:
    print("Caught Exception")

try:
    raise ImportError
except Exception:
    print("Caught ImportError via Exception")

try:
    raise ImportError
except ImportError:
    print("Caught ImportError")

try:
    raise IndentationError
except SyntaxError:
    print("Caught IndentationError via SyntaxError")

try:
    raise IndentationError
except IndentationError:
    print("Caught IndentationError")

try:
    raise IndexError
except LookupError:
    print("Caught IndexError via LookupError")

try:
    raise IndexError
except IndexError:
    print("Caught IndexError")

try:
    raise KeyError
except LookupError:
    print("Caught KeyError via LookupError")

try:
    raise KeyError
except KeyError:
    print("Caught KeyError")

try:
    raise LookupError
except Exception:
    print("Caught LookupError via Exception")

try:
    raise LookupError
except LookupError:
    print("Caught LookupError")



# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This script contains common functionality to assist a port in implementing
# make-pins.py, which is used to emit compile-time definitions of pin, AF, and
# ADC objects based on inputs from the vendor HAL/SDK and the board
# definition's pins.csv.

# The pins.csv file can contain empty lines, comments (a line beginning with "#")
# or pin definition lines.  Pin definition lines must be of the form:
#
#     board,cpu
#
# Where "board" is the user-facing name of the pin as specified by the particular
# board layout and markings, and "cpu" is the corresponding name of the CPU/MCU
# pin.
#
# The "board" entry may be absent if the CPU pin has no additional name, and both
# entries may start with "-" to hide them from the corresponding Python dict of
# pins, and hence hide them from the user (but they are still accessible in C).
#
# For example, take the following pins.csv file:
#
#     X1,PA0
#     -X2,PA1
#     X3,-PA2
#     -X4,-PA3
#     ,PA4
#     ,-PA5
#
# The first row here configures:
# - The CPU pin PA0 is labelled X1.
# - The Python user can access both by the names Pin("X1") and Pin("A0").
# - The Python user can access both by the members Pin.board.X1 and Pin.cpu.A0.
# - In C code they are available as pyb_pin_X1 and pin_A0.
#
# Prefixing the names with "-" hides them from the user.  The following table
# summarises the various possibilities:
#
#     pins.csv entry | board name | cpu name | C board name | C cpu name
#     ---------------+------------+----------+--------------+-----------
#     X1,PA0           "X1"         "A0"       pyb_pin_X1     pin_A0
#     -X2,PA1          -            "A1"       pyb_pin_X2     pin_A1
#     X3,-PA2          "X3"         -          pyb_pin_X3     pin_A2
#     -X4,-PA3         -            -          pyb_pin_X4     pin_A3
#     ,PA4             -            "A4"       -              pin_A4
#     ,-PA5            -            -          -              pin_A5

import argparse
import csv
import os
import sys


class PinGeneratorError(Exception):
    pass


# A port should define a subclass of Pin that knows how to validate cpu/board
# names and emits the required structures.
class Pin:
    def __init__(self, cpu_pin_name):
        self._cpu_pin_name = cpu_pin_name
        # Optional aliases for the board from pins.csv. Each entry is a tuple
        # of (name, hidden). Hidden board pins are in pins.csv with with a "-"
        # prefix and will available to C but not Python.
        self._board_pin_names = set()
        # An unavailable pin is one that is not explicitly mentioned at all in
        # pins.csv (or added explicitly with PinGenerator.add_cpu_pin).
        self._available = False
        # A hidden pin is one that is in pins.csv with a "-" prefix and will
        # be still available to C but not Python.
        self._hidden = False
        # Reference to the PinGenerator instance.
        self._generator = None

    # The name of the pin to use in MP_QSTR_{} or pin_{}. Defaults to the cpu name.
    def name(self):
        return self._cpu_pin_name

    # Add a board alias (e.g. from pins.csv).
    def add_board_pin_name(self, board_pin_name, hidden=False):
        self._board_pin_names.add(
            (
                board_pin_name,
                hidden,
            )
        )

    # Override this to handle an af specified in af.csv.
    def add_af(self, af_idx, af_name, af):
        raise NotImplementedError

    # Override this to verify that naming matches the MCU standard (e.g. "GPIOn" or "PXn").
    @staticmethod
    def validate_cpu_pin_name(cpu_pin_name):
        if not cpu_pin_name.strip():
            raise PinGeneratorError("Missing cpu pin name")

    # Override this to provide additional validation of board names.
    @staticmethod
    def validate_board_pin_name(board_pin_name):
        # TODO: ensure this is a valid Python/C identifier and can be used as MP_QSTR_foo.
        pass

    # Must be implemented when using NumericPinGenerator.
    # Returns the integer index, or None to exclude this pin from the table.
    def index(self):
        raise NotImplementedError

    # Can be overridden when using NumericPinGenerator.
    # Returns a string which is a C expression that evaluates to the index
    # e.g. `GPIO_NUM_7`.
    # This is used whenever the index is emitted in source code and defaults
    # to just returning the pin index as a literal.
    # Return None to exclude this pin from the table.
    def index_name(self):
        i = self.index()
        return str(i) if i is not None else None

    # Returns an expression that defines the pin. e.g. `{ .base { ... }, .x }`.
    # This is used as the RHS of the `const machine_pin_obj_t
    # pin_EXT_GPIO0_obj =` statements for named pins, and elements of
    # `machine_pin_obj_table` for numeric pins.
    # This will typically might be implemented as an invocation of a macro
    # defined in the port-specific prefix.
    def definition(self):
        raise NotImplementedError

    # Whether the pin object should be declared as "const". This should be True
    # for most pins, but some special cases (e.g. external pins on rp2) might
    # need mutable pin objects (e.g. to track current pin state).
    def is_const(self):
        return True

    # Optionally return a preprocessor expression that can be used to suppress
    # this pin (e.g. `MICROPY_HW_ENABLE_GPIOn`).
    def enable_macro(self):
        return None

    # Override this to output any additional per-pin definitions or other
    # content that should appear at the start of the source output.
    # This could be used to define additional objects such as IRQs or AFs.
    def print_source(self, out_source):
        pass


# A port should define a subclass of PinGenerator (or NumericPinGenerator).
class PinGenerator:
    def __init__(self, pin_type, enable_af=False):
        self._pins = []
        self._pin_type = pin_type
        self._enable_af = enable_af

    # Allows a port to define a known cpu pin (without relying on it being in the
    # csv file).
    def add_cpu_pin(self, cpu_pin_name, available=True):
        pin = self._pin_type(cpu_pin_name)
        pin._available = available
        self._pins.append(pin)
        pin._generator = self
        return pin

    # Iterate just the available pins (i.e. ones in pins.csv).
    def available_pins(self, exclude_hidden=False):
        for pin in self._pins:
            if not pin._available:
                continue
            if exclude_hidden and pin._hidden:
                continue
            yield pin

    # Allows a port to add additional command-line arguments to be handled.
    def extra_args(self, parser):
        pass

    # Load board->cpu mapping from csv.
    def parse_board_csv(self, filename):
        with open(filename, "r") as csvfile:
            rows = csv.reader(csvfile)
            for linenum, row in enumerate(rows):
                try:
                    # Skip empty lines, and lines starting with "#".
                    if len(row) == 0 or row[0].startswith("#"):
                        continue

                    # Lines must be pairs of names.
                    if len(row) != 2:
                        raise PinGeneratorError("Expecting two entries in each row")
                    board_pin_name, cpu_pin_name = (x.strip() for x in row)

                    # All rows must include a cpu name.
                    cpu_hidden = False
                    if cpu_pin_name.startswith("-"):
                        cpu_hidden = True
                        cpu_pin_name = cpu_pin_name[1:]
                    self._pin_type.validate_cpu_pin_name(cpu_pin_name)
                    pin = self.find_pin_by_cpu_pin_name(cpu_pin_name, create=True)
                    pin._available = True  # It's in pins.csv so must be available.
                    pin._hidden = cpu_hidden  # Optionally don't make available to Python.

                    # Rows can optionally alias to a board name.
                    if board_pin_name:
                        board_hidden = False
                        if board_pin_name.startswith("-"):
                            board_hidden = True
                            board_pin_name = board_pin_name[1:]
                        self._pin_type.validate_board_pin_name(board_pin_name)
                        pin.add_board_pin_name(board_pin_name, board_hidden)

                # Inject "file:line: " into the exception.
                except PinGeneratorError as er:
                    raise PinGeneratorError("{}:{}: {}".format(filename, linenum, er))

    def parse_af_csv(self, filename, header_rows=1, pin_col=0, af_col=1):
        headings = {}
        with open(filename, "r") as csvfile:
            rows = csv.reader(csvfile)
            for linenum, row in enumerate(rows):
                try:
                    # Skip empty lines, and lines starting with "#".
                    if len(row) == 0 or row[0].startswith("#"):
                        continue

                    # Consume `header_rows` non-blank/comment rows at the start.
                    if header_rows:
                        if not headings:
                            # If this is the first header row then initialise
                            # the headings dict.
                            for af_idx, header in enumerate(row[af_col:]):
                                headings[af_idx] = header.strip()
                        header_rows -= 1
                        continue

                    # Lines must be pairs of names.
                    if len(row) <= max(pin_col, af_col):
                        raise PinGeneratorError(
                            "Expecting {} entries in each row".format(max(pin_col, af_col))
                        )

                    cpu_pin_name = row[pin_col].strip()
                    if cpu_pin_name == "-":
                        continue
                    self._pin_type.validate_cpu_pin_name(cpu_pin_name)
                    pin = self.find_pin_by_cpu_pin_name(cpu_pin_name, create=True)

                    for af_idx, af in enumerate(row[af_col:]):
                        af = af.strip()
                        if not af:
                            continue
                        pin.add_af(af_idx, headings.get(af_idx, ""), af)

                # Inject "file:line: " into the exception.
                except PinGeneratorError as er:
                    raise PinGeneratorError("{}:{}: {}".format(filename, linenum, er))

    # Find an existing pin.
    def find_pin_by_cpu_pin_name(self, cpu_pin_name, create=True):
        for pin in self._pins:
            if pin._cpu_pin_name == cpu_pin_name:
                return pin
        if create:
            return self.add_cpu_pin(cpu_pin_name, available=False)
        else:
            raise PinGeneratorError("Unknown cpu pin {}".format(cpu_pin_name))

    # Print the locals dict for Pin.board.
    def print_board_locals_dict(self, out_source):
        print(file=out_source)
        print(
            "static const mp_rom_map_elem_t machine_pin_board_pins_locals_dict_table[] = {",
            file=out_source,
        )
        for pin in self.available_pins():
            for board_pin_name, board_hidden in pin._board_pin_names:
                if board_hidden:
                    # Don't include hidden pins in Pins.board.
                    continue

                # We don't use the enable macro for board pins, because they
                # shouldn't be referenced in pins.csv unless they're
                # available.
                print(
                    "    {{ MP_ROM_QSTR(MP_QSTR_{:s}), MP_ROM_PTR(pin_{:s}) }},".format(
                        board_pin_name,
                        pin.name(),
                    ),
                    file=out_source,
                )
        print("};", file=out_source)
        print(
            "MP_DEFINE_CONST_DICT(machine_pin_board_pins_locals_dict, machine_pin_board_pins_locals_dict_table);",
            file=out_source,
        )

    # Print the locals dict for Pin.cpu.
    def print_cpu_locals_dict(self, out_source):
        print(file=out_source)
        print(
            "static const mp_rom_map_elem_t machine_pin_cpu_pins_locals_dict_table[] = {",
            file=out_source,
        )
        for pin in self.available_pins(exclude_hidden=True):
            m = pin.enable_macro()
            if m:
                print("    #if {}".format(m), file=out_source)
            print(
                "    {{ MP_ROM_QSTR(MP_QSTR_{:s}), MP_ROM_PTR(pin_{:s}) }},".format(
                    pin.name(),
                    pin.name(),
                ),
                file=out_source,
            )
            if m:
                print("    #endif", file=out_source)
        print("};", file=out_source)
        print(
            "MP_DEFINE_CONST_DICT(machine_pin_cpu_pins_locals_dict, machine_pin_cpu_pins_locals_dict_table);",
            file=out_source,
        )

    # NumericPinGenerator can override this to use an entry in machine_pin_obj_table.
    def _cpu_pin_pointer(self, pin):
        return "&pin_{:s}_obj".format(pin.name())

    # Allow a port to prefix the board pin macro names with something.
    # e.g. STM32 does pyb_pin_NAME whereas other ports using pin_NAME.
    def board_name_define_prefix(self):
        return ""

    # Print the pin_CPUNAME and pin_BOARDNAME macros.
    def print_defines(self, out_header, cpu=True, board=True):
        # Provide #defines for each cpu pin.
        for pin in self.available_pins():
            print(file=out_header)
            m = pin.enable_macro()
            if m:
                print("#if {}".format(m), file=out_header)

            # #define pin_CPUNAME (...)
            if cpu:
                print(
                    "#define pin_{:s} ({:s})".format(pin.name(), self._cpu_pin_pointer(pin)),
                    file=out_header,
                )

            # #define pin_BOARDNAME (pin_CPUNAME)
            if board:
                for board_pin_name, _board_hidden in pin._board_pin_names:
                    # Note: Hidden board pins are still available to C via the macro.
                    # Note: The RHS isn't wrapped in (), which is necessary to make the
                    # STATIC_AF_ macro work on STM32.
                    print(
                        "#define {:s}pin_{:s} pin_{:s}".format(
                            self.board_name_define_prefix(),
                            board_pin_name,
                            pin.name(),
                        ),
                        file=out_header,
                    )

            if m:
                print("#endif", file=out_header)

    def print_pin_objects(self, out_source):
        print(file=out_source)
        for pin in self.available_pins():
            m = pin.enable_macro()
            if m:
                print("#if {}".format(m), file=out_source)
            print(
                "{:s}machine_pin_obj_t pin_{:s}_obj = {:s};".format(
                    "const " if pin.is_const() else "",
                    pin.name(),
                    pin.definition(),
                ),
                file=out_source,
            )
            if m:
                print("#endif", file=out_source)

    def print_pin_object_externs(self, out_header):
        print(file=out_header)
        for pin in self.available_pins():
            m = pin.enable_macro()
            if m:
                print("#if {}".format(m), file=out_header)
            print(
                "extern {:s}machine_pin_obj_t pin_{:s}_obj;".format(
                    "const " if pin.is_const() else "",
                    pin.name(),
                ),
                file=out_header,
            )
            if m:
                print("#endif", file=out_header)

    def print_source(self, out_source):
        self.print_pin_objects(out_source)
        self.print_cpu_locals_dict(out_source)
        self.print_board_locals_dict(out_source)

    def print_header(self, out_header):
        self.print_pin_object_externs(out_header)
        self.print_defines(out_header)

    # A port can override this if it has extra input files (e.g. af.csv) to load.
    def load_inputs(self, out_source):
        # Optionally load pins.csv to get cpu->board name mappings.
        if self._enable_af and self.args.af_csv:
            print("// --af-csv {:s}".format(self.args.af_csv), file=out_source)
            self.parse_af_csv(self.args.af_csv)

        # Optionally load pins.csv to get cpu->board name mappings.
        if self.args.board_csv:
            print("// --board-csv {:s}".format(self.args.board_csv), file=out_source)
            self.parse_board_csv(self.args.board_csv)

        # Prepend the prefix file to the start of the output.
        if self.args.prefix:
            print("// --prefix {:s}".format(self.args.prefix), file=out_source)
            print(file=out_source)
            with open(self.args.prefix, "r") as prefix_file:
                print(prefix_file.read(), end="", file=out_source)

    # A port can override this to do extra work after the main source+header
    # have been written, such as generating additional header files.
    def generate_extra_files(self):
        pass

    def main(self):
        parser = argparse.ArgumentParser(description="Generate board specific pin file")
        parser.add_argument("--board-csv")
        if self._enable_af:
            parser.add_argument("--af-csv")
        parser.add_argument("--prefix")
        parser.add_argument("--output-source")
        parser.add_argument("--output-header")
        self.extra_args(parser)
        self.args = parser.parse_args()

        try:
            with open(self.args.output_source, "w") as out_source:
                print("// This file was automatically generated by make-pins.py", file=out_source)
                print("//", file=out_source)

                # Load additional files (including port-specific ones).
                self.load_inputs(out_source)

                # Allow a port to print arbitrary per-pin content.
                for pin in self.available_pins():
                    pin.print_source(out_source)

                # Print the tables and dictionaries.
                self.print_source(out_source)

            with open(self.args.output_header, "w") as out_header:
                self.print_header(out_header)

            self.generate_extra_files()
        except PinGeneratorError as er:
            print(er)
            sys.exit(1)


# For ports that use numeric pin identifiers (e.g. ESP32, rp2).
# This emits the machine_pin_obj_t instances as an array (machine_pin_obj_table).
class NumericPinGenerator(PinGenerator):
    # This should return a const expression that is the number of GPIO pins
    # for this board.
    def cpu_table_size(self):
        raise NotImplementedError

    def print_cpu_table(self, out_source):
        # Print machine_pin_obj_table, where each element is `[n] = {obj}`.
        print(file=out_source)
        print(
            "const machine_pin_obj_t machine_pin_obj_table[{}] = {{".format(self.cpu_table_size()),
            file=out_source,
        )
        for pin in self.available_pins():
            n = pin.index_name()
            if n is None:
                continue

            m = pin.enable_macro()
            if m:
                print("    #if {}".format(m), file=out_source)
            print(
                "    [{:s}] = {:s},".format(
                    pin.index_name(),
                    pin.definition(),
                ),
                file=out_source,
            )
            if m:
                print("    #endif", file=out_source)
        print("};", file=out_source)

        # For pins that do not have an index, print them out in the same style as PinGenerator.
        print(file=out_source)
        for pin in self.available_pins():
            n = pin.index_name()
            if n is not None:
                continue

            m = pin.enable_macro()
            if m:
                print("#if {}".format(m), file=out_source)
            print(
                "{:s}machine_pin_obj_t pin_{:s}_obj = {:s};".format(
                    "const " if pin.is_const() else "",
                    pin.name(),
                    pin.definition(),
                ),
                file=out_source,
            )
            if m:
                print("#endif", file=out_source)

    # Replace PinGenerator's implementation to print the numeric table.
    def print_source(self, out_source):
        self.print_cpu_table(out_source)
        self.print_board_locals_dict(out_source)

    # Replace PinGenerator's implementation to only print the defines.
    def print_header(self, out_header):
        self.print_defines(out_header)

    def _cpu_pin_pointer(self, pin):
        n = pin.index_name()
        if n is not None:
            return "&machine_pin_obj_table[{:s}]".format(pin.index_name())
        else:
            return super()._cpu_pin_pointer(pin)



print(id(1) == id(2))
print(id(None) == id(None))
# This can't be true per Python semantics, just CPython implementation detail
#print(id([]) == id([]))

l = [1, 2]
print(id(l) == id(l))

f = lambda:None
print(id(f) == id(f))



import esp


class FlashBdev:
    SEC_SIZE = 4096

    def __init__(self, start_sec, blocks):
        self.start_sec = start_sec
        self.blocks = blocks

    def readblocks(self, n, buf, off=0):
        # print("readblocks(%s, %x(%d), %d)" % (n, id(buf), len(buf), off))
        esp.flash_read((n + self.start_sec) * self.SEC_SIZE + off, buf)

    def writeblocks(self, n, buf, off=None):
        # print("writeblocks(%s, %x(%d), %d)" % (n, id(buf), len(buf), off))
        # assert len(buf) <= self.SEC_SIZE, len(buf)
        if off is None:
            esp.flash_erase(n + self.start_sec)
            off = 0
        esp.flash_write((n + self.start_sec) * self.SEC_SIZE + off, buf)

    def ioctl(self, op, arg):
        # print("ioctl(%d, %r)" % (op, arg))
        if op == 4:  # MP_BLOCKDEV_IOCTL_BLOCK_COUNT
            return self.blocks
        if op == 5:  # MP_BLOCKDEV_IOCTL_BLOCK_SIZE
            return self.SEC_SIZE
        if op == 6:  # MP_BLOCKDEV_IOCTL_BLOCK_ERASE
            esp.flash_erase(arg + self.start_sec)
            return 0


size = esp.flash_size()
if size < 1024 * 1024:
    bdev = None
else:
    start_sec = esp.flash_user_start() // FlashBdev.SEC_SIZE
    if start_sec < 256:
        start_sec += 1  # Reserve space for native code
    # 20K at the flash end is reserved for SDK params storage
    bdev = FlashBdev(start_sec, (size - 20480) // FlashBdev.SEC_SIZE - start_sec)



# MicroPython asyncio module, for use with webassembly port
# MIT license; Copyright (c) 2019-2024 Damien P. George

from time import ticks_ms as ticks, ticks_diff, ticks_add
import sys, js, jsffi

# Import TaskQueue and Task from built-in C code.
from _asyncio import TaskQueue, Task


################################################################################
# Exceptions


class CancelledError(BaseException):
    pass


class TimeoutError(Exception):
    pass


# Used when calling Loop.call_exception_handler.
_exc_context = {"message": "Task exception wasn't retrieved", "exception": None, "future": None}


################################################################################
# Sleep functions


# "Yield" once, then raise StopIteration
class SingletonGenerator:
    def __init__(self):
        self.state = None
        self.exc = StopIteration()

    def __iter__(self):
        return self

    def __next__(self):
        if self.state is not None:
            _task_queue.push(cur_task, self.state)
            self.state = None
            return None
        else:
            self.exc.__traceback__ = None
            raise self.exc


# Pause task execution for the given time (integer in milliseconds, uPy extension)
# Use a SingletonGenerator to do it without allocating on the heap
def sleep_ms(t, sgen=SingletonGenerator()):
    assert sgen.state is None
    sgen.state = ticks_add(ticks(), max(0, t))
    return sgen


# Pause task execution for the given time (in seconds)
def sleep(t):
    return sleep_ms(int(t * 1000))


################################################################################
# Main run loop

asyncio_timer = None


class TopLevelCoro:
    @staticmethod
    def set(resolve, reject):
        TopLevelCoro.resolve = resolve
        TopLevelCoro.reject = reject

    @staticmethod
    def send(value):
        TopLevelCoro.resolve()


class ThenableEvent:
    def __init__(self, thenable):
        self.result = None  # Result of the thenable
        self.waiting = None  # Task waiting on completion of this thenable
        thenable.then(self.set)

    def set(self, value=None):
        # Thenable/Promise is fulfilled, set result and schedule any waiting task.
        self.result = value
        if self.waiting:
            _task_queue.push(self.waiting)
            self.waiting = None

    def remove(self, task):
        self.waiting = None

    # async
    def wait(self):
        # Set the calling task as the task waiting on this thenable.
        self.waiting = cur_task
        # Set calling task's data to this object so it can be removed if needed.
        cur_task.data = self
        # Wait for the thenable to fulfill.
        yield
        # Return the result of the thenable.
        return self.result


# Ensure the awaitable is a task
def _promote_to_task(aw):
    return aw if isinstance(aw, Task) else create_task(aw)


def _schedule_run_iter(dt):
    global asyncio_timer
    if asyncio_timer is not None:
        js.clearTimeout(asyncio_timer)
    asyncio_timer = js.setTimeout(_run_iter, dt)


def _run_iter():
    global cur_task
    excs_all = (CancelledError, Exception)  # To prevent heap allocation in loop
    excs_stop = (CancelledError, StopIteration)  # To prevent heap allocation in loop
    while True:
        # Wait until the head of _task_queue is ready to run
        t = _task_queue.peek()
        if t:
            # A task waiting on _task_queue; "ph_key" is time to schedule task at
            dt = max(0, ticks_diff(t.ph_key, ticks()))
        else:
            # No tasks can be woken so finished running
            cur_task = _top_level_task
            return

        if dt > 0:
            # schedule to call again later
            cur_task = _top_level_task
            _schedule_run_iter(dt)
            return

        # Get next task to run and continue it
        t = _task_queue.pop()
        cur_task = t
        try:
            # Continue running the coroutine, it's responsible for rescheduling itself
            exc = t.data
            if not exc:
                t.coro.send(None)
            else:
                # If the task is finished and on the run queue and gets here, then it
                # had an exception and was not await'ed on.  Throwing into it now will
                # raise StopIteration and the code below will catch this and run the
                # call_exception_handler function.
                t.data = None
                t.coro.throw(exc)
        except excs_all as er:
            # Check the task is not on any event queue
            assert t.data is None
            # This task is done.
            if t.state:
                # Task was running but is now finished.
                waiting = False
                if t.state is True:
                    # "None" indicates that the task is complete and not await'ed on (yet).
                    t.state = None
                elif callable(t.state):
                    # The task has a callback registered to be called on completion.
                    t.state(t, er)
                    t.state = False
                    waiting = True
                else:
                    # Schedule any other tasks waiting on the completion of this task.
                    while t.state.peek():
                        _task_queue.push(t.state.pop())
                        waiting = True
                    # "False" indicates that the task is complete and has been await'ed on.
                    t.state = False
                if not waiting and not isinstance(er, excs_stop):
                    # An exception ended this detached task, so queue it for later
                    # execution to handle the uncaught exception if no other task retrieves
                    # the exception in the meantime (this is handled by Task.throw).
                    _task_queue.push(t)
                # Save return value of coro to pass up to caller.
                t.data = er
            elif t.state is None:
                # Task is already finished and nothing await'ed on the task,
                # so call the exception handler.

                # Save exception raised by the coro for later use.
                t.data = exc

                # Create exception context and call the exception handler.
                _exc_context["exception"] = exc
                _exc_context["future"] = t
                Loop.call_exception_handler(_exc_context)


# Create and schedule a new task from a coroutine.
def create_task(coro):
    if not hasattr(coro, "send"):
        raise TypeError("coroutine expected")
    t = Task(coro, globals())
    _task_queue.push(t)
    return t


# Task used to suspend and resume top-level await.
_top_level_task = Task(TopLevelCoro, globals())

################################################################################
# Event loop wrapper


cur_task = _top_level_task


class Loop:
    _exc_handler = None

    def create_task(coro):
        return create_task(coro)

    def close():
        pass

    def set_exception_handler(handler):
        Loop._exc_handler = handler

    def get_exception_handler():
        return Loop._exc_handler

    def default_exception_handler(loop, context):
        print(context["message"], file=sys.stderr)
        print("future:", context["future"], "coro=", context["future"].coro, file=sys.stderr)
        sys.print_exception(context["exception"], sys.stderr)

    def call_exception_handler(context):
        (Loop._exc_handler or Loop.default_exception_handler)(Loop, context)


def get_event_loop():
    return Loop


def current_task():
    if cur_task is None:
        raise RuntimeError("no running event loop")
    return cur_task


def new_event_loop():
    global _task_queue
    _task_queue = TaskQueue(_schedule_run_iter)  # TaskQueue of Task instances.
    return Loop


# Initialise default event loop.
new_event_loop()



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2023 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import ast, hashlib, os, sys
from collections import namedtuple


def stdout_write_bytes(b):
    b = b.replace(b"\x04", b"")
    if hasattr(sys.stdout, "buffer"):
        sys.stdout.buffer.write(b)
        sys.stdout.buffer.flush()
    else:
        text = b.decode(sys.stdout.encoding, "strict")
        sys.stdout.write(text)


class TransportError(Exception):
    pass


class TransportExecError(TransportError):
    def __init__(self, status_code, error_output):
        self.status_code = status_code
        self.error_output = error_output
        super().__init__(error_output)


listdir_result = namedtuple("dir_result", ["name", "st_mode", "st_ino", "st_size"])


# Takes a Transport error (containing the text of an OSError traceback) and
# raises it as the corresponding OSError-derived exception.
def _convert_filesystem_error(e, info):
    if "OSError" in e.error_output and "ENOENT" in e.error_output:
        return FileNotFoundError(info)
    if "OSError" in e.error_output and "EISDIR" in e.error_output:
        return IsADirectoryError(info)
    if "OSError" in e.error_output and "EEXIST" in e.error_output:
        return FileExistsError(info)
    if "OSError" in e.error_output and "ENODEV" in e.error_output:
        return FileNotFoundError(info)
    return e


class Transport:
    def fs_listdir(self, src=""):
        buf = bytearray()

        def repr_consumer(b):
            buf.extend(b.replace(b"\x04", b""))

        cmd = "import os\nfor f in os.ilistdir(%s):\n" " print(repr(f), end=',')" % (
            ("'%s'" % src) if src else ""
        )
        try:
            buf.extend(b"[")
            self.exec(cmd, data_consumer=repr_consumer)
            buf.extend(b"]")
        except TransportExecError as e:
            raise _convert_filesystem_error(e, src) from None

        return [
            listdir_result(*f) if len(f) == 4 else listdir_result(*(f + (0,)))
            for f in ast.literal_eval(buf.decode())
        ]

    def fs_stat(self, src):
        try:
            self.exec("import os")
            return os.stat_result(self.eval("os.stat(%s)" % ("'%s'" % src)))
        except TransportExecError as e:
            raise _convert_filesystem_error(e, src) from None

    def fs_exists(self, src):
        try:
            self.fs_stat(src)
            return True
        except OSError:
            return False

    def fs_isdir(self, src):
        try:
            mode = self.fs_stat(src).st_mode
            return (mode & 0x4000) != 0
        except OSError:
            # Match CPython, a non-existent path is not a directory.
            return False

    def fs_printfile(self, src, chunk_size=256):
        cmd = (
            "with open('%s') as f:\n while 1:\n"
            "  b=f.read(%u)\n  if not b:break\n  print(b,end='')" % (src, chunk_size)
        )
        try:
            self.exec(cmd, data_consumer=stdout_write_bytes)
        except TransportExecError as e:
            raise _convert_filesystem_error(e, src) from None

    def fs_readfile(self, src, chunk_size=256, progress_callback=None):
        if progress_callback:
            src_size = self.fs_stat(src).st_size

        contents = bytearray()

        try:
            self.exec("f=open('%s','rb')\nr=f.read" % src)
            while True:
                chunk = self.eval("r({})".format(chunk_size))
                if not chunk:
                    break
                contents.extend(chunk)
                if progress_callback:
                    progress_callback(len(contents), src_size)
            self.exec("f.close()")
        except TransportExecError as e:
            raise _convert_filesystem_error(e, src) from None

        return contents

    def fs_writefile(self, dest, data, chunk_size=256, progress_callback=None):
        if progress_callback:
            src_size = len(data)
            written = 0

        try:
            self.exec("f=open('%s','wb')\nw=f.write" % dest)
            while data:
                chunk = data[:chunk_size]
                self.exec("w(" + repr(chunk) + ")")
                data = data[len(chunk) :]
                if progress_callback:
                    written += len(chunk)
                    progress_callback(written, src_size)
            self.exec("f.close()")
        except TransportExecError as e:
            raise _convert_filesystem_error(e, dest) from None

    def fs_mkdir(self, path):
        try:
            self.exec("import os\nos.mkdir('%s')" % path)
        except TransportExecError as e:
            raise _convert_filesystem_error(e, path) from None

    def fs_rmdir(self, path):
        try:
            self.exec("import os\nos.rmdir('%s')" % path)
        except TransportExecError as e:
            raise _convert_filesystem_error(e, path) from None

    def fs_rmfile(self, path):
        try:
            self.exec("import os\nos.remove('%s')" % path)
        except TransportExecError as e:
            raise _convert_filesystem_error(e, path) from None

    def fs_touchfile(self, path):
        try:
            self.exec("f=open('%s','a')\nf.close()" % path)
        except TransportExecError as e:
            raise _convert_filesystem_error(e, path) from None

    def fs_hashfile(self, path, algo, chunk_size=256):
        try:
            self.exec("import hashlib\nh = hashlib.{algo}()".format(algo=algo))
        except TransportExecError:
            # hashlib (or hashlib.{algo}) not available on device. Do the hash locally.
            data = self.fs_readfile(path, chunk_size=chunk_size)
            return getattr(hashlib, algo)(data).digest()
        try:
            self.exec(
                "buf = memoryview(bytearray({chunk_size}))\nwith open('{path}', 'rb') as f:\n while True:\n  n = f.readinto(buf)\n  if n == 0:\n   break\n  h.update(buf if n == {chunk_size} else buf[:n])\n".format(
                    chunk_size=chunk_size, path=path
                )
            )
            return self.eval("h.digest()")
        except TransportExecError as e:
            raise _convert_filesystem_error(e, path) from None



import sys

# Depending on the port, the numbers in this test may be implemented as "small"
# native 64 bit ints, arbitrary precision large ints, or large integers using 64-bit
# long longs.

try:
    x = int.from_bytes(b"\x6F\xAB\xCD\x12\x34\x56\x78\xFB", "big")
except OverflowError:
    print("SKIP")  # Port can't represent this size of integer at all
    raise SystemExit

print(hex(x))
b = x.to_bytes(8, "little")
print(b)
print(x.to_bytes(8, "big"))

# padding in output
print(x.to_bytes(20, "little"))
print(x.to_bytes(20, "big"))

# check that extra zero bytes don't change the internal int value
print(int.from_bytes(b + bytes(10), "little") == x)

# can't write to a zero-length bytes object
try:
    x.to_bytes(0, "little")
except OverflowError:
    print("OverflowError")

# or one that it too short
try:
    x.to_bytes(7, "big")
except OverflowError:
    print("OverflowError")

# negative representations

# MicroPython int.to_bytes() behaves as if signed=True for negative numbers
if "micropython" in repr(sys.implementation):

    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e)
else:
    # Implement MicroPython compatible behaviour for CPython
    def to_bytes_compat(i, l, e):
        return i.to_bytes(l, e, signed=i < 0)


print(to_bytes_compat(-x, 8, "little"))
print(to_bytes_compat(-x, 20, "big"))
print(to_bytes_compat(-x, 20, "little"))



# Test network loopback behaviour

try:
    import asyncio
except ImportError:
    print("SKIP")
    raise SystemExit


async def client(host, port):
    print(f"client open_connection to {host}:{port}")
    reader, writer = await asyncio.open_connection(host, port)

    data_in = b"A" * 100

    print("client writing")
    writer.write(data_in)
    await writer.drain()

    await asyncio.sleep(0.1)

    print("client reading")
    data = await reader.readexactly(100)
    print(f"client got {len(data)} bytes")

    assert data_in == data

    print("client closing")

    writer.close()
    await writer.wait_closed()

    print("client closed")


async def echo_handler(reader, writer):
    print("handler reading")
    await asyncio.sleep(0.1)
    data = await reader.readexactly(100)
    print(f"handler got {len(data)} bytes")

    print("handler writing")
    writer.write(data)
    await writer.drain()

    print("handler closing")

    writer.close()
    await writer.wait_closed()

    print("handler closed")


async def test(host, port):
    print(f"create server on {host}:{port}")
    server = await asyncio.start_server(echo_handler, host, port)

    async with server:
        print("server started")
        await client("127.0.0.1", 8080)
    print("server closed")


asyncio.run(test("0.0.0.0", 8080))



# Exception chaining is not supported, but check that basic
# exception works as expected.

try:
    raise Exception from None
except Exception:
    print("Caught Exception")

try:
    try:
        raise ValueError("Value")
    except Exception as exc:
        raise RuntimeError("Runtime") from exc
except Exception as ex2:
    print("Caught Exception:", ex2)



# OMGS3 Helper Library
# MIT license; Copyright (c) 2024 Seon Rozenblum - Unexpected Maker
#
# Project home:
#   https://omgs3.io

# Import required libraries
from micropython import const
from machine import Pin, I2C
from max17048 import MAX17048
import time

# Initialize I2C bus
fg_i2c = I2C(0, scl=Pin.board.I2C_SCL, sda=Pin.board.I2C_SDA)

# Create an instance of the MAX17048 class
max17048 = MAX17048(fg_i2c)


# Helper functions
def get_bat_voltage():
    """Read the battery voltage from the fuel gauge"""
    voltage = max17048.cell_voltage
    print(f"Bat Voltage: {voltage}V")
    return voltage


def get_state_of_charge():
    """Read the battery state of charge from the fuel gauge"""
    soc = max17048.state_of_charge
    print(f"State of Charge: {soc}%")
    return soc


def get_vbus_present():
    """Detect if VBUS (5V) power source is present"""
    return Pin(Pin.board.VBUS_SENSE, Pin.IN).value() == 1


def set_pixel_power(state):
    """Enable or Disable power to the onboard NeoPixel to either show colour, or to reduce power for deep sleep."""
    Pin(Pin.board.RGB_PWR, Pin.OUT).value(state)


# NeoPixel rainbow colour wheel
def rgb_color_wheel(wheel_pos):
    """Color wheel to allow for cycling through the rainbow of RGB colors."""
    wheel_pos = wheel_pos % 255

    if wheel_pos < 85:
        return 255 - wheel_pos * 3, 0, wheel_pos * 3
    elif wheel_pos < 170:
        wheel_pos -= 85
        return 0, wheel_pos * 3, 255 - wheel_pos * 3
    else:
        wheel_pos -= 170
        return wheel_pos * 3, 255 - wheel_pos * 3, 0



import time
from signal import *


print("org signal() val:", signal(SIGINT, SIG_IGN))
print("read back signal() val:", signal(SIGINT, SIG_IGN))

print("You can't break me - use SIGTERM")
while 1:
    time.sleep(1)



#!/usr/bin/python

# Written by Antonio Galea - 2010/11/18
# Distributed under Gnu LGPL 3.0
# see http://www.gnu.org/licenses/lgpl-3.0.txt

import sys, struct, zlib, os
from optparse import OptionParser

DEFAULT_DEVICE = "0x0483:0xdf11"


def named(tuple, names):
    return dict(zip(names.split(), tuple))


def consume(fmt, data, names):
    n = struct.calcsize(fmt)
    return named(struct.unpack(fmt, data[:n]), names), data[n:]


def cstring(string):
    return string.split(b"\0", 1)[0]


def compute_crc(data):
    return 0xFFFFFFFF & -zlib.crc32(data) - 1


def parse(file, dump_images=False):
    print('File: "%s"' % file)
    data = open(file, "rb").read()
    crc = compute_crc(data[:-4])
    prefix, data = consume("<5sBIB", data, "signature version size targets")
    print("%(signature)s v%(version)d, image size: %(size)d, targets: %(targets)d" % prefix)
    for t in range(prefix["targets"]):
        tprefix, data = consume(
            "<6sBI255s2I", data, "signature altsetting named name size elements"
        )
        tprefix["num"] = t
        if tprefix["named"]:
            tprefix["name"] = cstring(tprefix["name"])
        else:
            tprefix["name"] = ""
        print(
            '%(signature)s %(num)d, alt setting: %(altsetting)s, name: "%(name)s", size: %(size)d, elements: %(elements)d'
            % tprefix
        )
        tsize = tprefix["size"]
        target, data = data[:tsize], data[tsize:]
        for e in range(tprefix["elements"]):
            eprefix, target = consume("<2I", target, "address size")
            eprefix["num"] = e
            print("  %(num)d, address: 0x%(address)08x, size: %(size)d" % eprefix)
            esize = eprefix["size"]
            image, target = target[:esize], target[esize:]
            if dump_images:
                out = "%s.target%d.image%d.bin" % (file, t, e)
                open(out, "wb").write(image)
                print('    DUMPED IMAGE TO "%s"' % out)
        if len(target):
            print("target %d: PARSE ERROR" % t)
    suffix = named(struct.unpack("<4H3sBI", data[:16]), "device product vendor dfu ufd len crc")
    print(
        "usb: %(vendor)04x:%(product)04x, device: 0x%(device)04x, dfu: 0x%(dfu)04x, %(ufd)s, %(len)d, 0x%(crc)08x"
        % suffix
    )
    if crc != suffix["crc"]:
        print("CRC ERROR: computed crc32 is 0x%08x" % crc)
    data = data[16:]
    if data:
        print("PARSE ERROR")


def build(file, targets, device=DEFAULT_DEVICE):
    data = b""
    for t, target in enumerate(targets):
        tdata = b""
        for image in target:
            # pad image to 8 bytes (needed at least for L476)
            pad = (8 - len(image["data"]) % 8) % 8
            image["data"] = image["data"] + bytes(bytearray(8)[0:pad])
            #
            tdata += struct.pack("<2I", image["address"], len(image["data"])) + image["data"]
        tdata = (
            struct.pack("<6sBI255s2I", b"Target", 0, 1, b"ST...", len(tdata), len(target)) + tdata
        )
        data += tdata
    data = struct.pack("<5sBIB", b"DfuSe", 1, len(data) + 11, len(targets)) + data
    v, d = map(lambda x: int(x, 0) & 0xFFFF, device.split(":", 1))
    data += struct.pack("<4H3sB", 0, d, v, 0x011A, b"UFD", 16)
    crc = compute_crc(data)
    data += struct.pack("<I", crc)
    open(file, "wb").write(data)


if __name__ == "__main__":
    usage = """
%prog [-d|--dump] infile.dfu
%prog {-b|--build} address:file.bin [-b address:file.bin ...] [{-D|--device}=vendor:device] outfile.dfu"""
    parser = OptionParser(usage=usage)
    parser.add_option(
        "-b",
        "--build",
        action="append",
        dest="binfiles",
        help="build a DFU file from given BINFILES",
        metavar="BINFILES",
    )
    parser.add_option(
        "-D",
        "--device",
        action="store",
        dest="device",
        help="build for DEVICE, defaults to %s" % DEFAULT_DEVICE,
        metavar="DEVICE",
    )
    parser.add_option(
        "-d",
        "--dump",
        action="store_true",
        dest="dump_images",
        default=False,
        help="dump contained images to current directory",
    )
    (options, args) = parser.parse_args()

    if options.binfiles and len(args) == 1:
        target = []
        for arg in options.binfiles:
            try:
                address, binfile = arg.split(":", 1)
            except ValueError:
                print("Address:file couple '%s' invalid." % arg)
                sys.exit(1)
            try:
                address = int(address, 0) & 0xFFFFFFFF
            except ValueError:
                print("Address %s invalid." % address)
                sys.exit(1)
            if not os.path.isfile(binfile):
                print("Unreadable file '%s'." % binfile)
                sys.exit(1)
            target.append({"address": address, "data": open(binfile, "rb").read()})
        outfile = args[0]
        device = DEFAULT_DEVICE
        if options.device:
            device = options.device
        try:
            v, d = map(lambda x: int(x, 0) & 0xFFFF, device.split(":", 1))
        except:
            print("Invalid device '%s'." % device)
            sys.exit(1)
        build(outfile, [target], device)
    elif len(args) == 1:
        infile = args[0]
        if not os.path.isfile(infile):
            print("Unreadable file '%s'." % infile)
            sys.exit(1)
        parse(infile, dump_images=options.dump_images)
    else:
        parser.print_help()
        sys.exit(1)



try:
    import btree
    import io
except ImportError:
    print("SKIP")
    raise SystemExit

f = io.BytesIO()
db = btree.open(f)

db[b"foo"] = b"42"

db.close()

# Accessing an already-closed database should fail.
try:
    print(db[b"foo"])
except ValueError:
    print("ValueError")

try:
    db[b"bar"] = b"43"
except ValueError:
    print("ValueError")

try:
    db.flush()
except ValueError:
    print("ValueError")

try:
    for k, v in db.items():
        pass
except ValueError:
    print("ValueError")

# Closing and printing an already-closed database should not fail.
db.close()
print(db)



#!/usr/bin/env python3
#
# This file is part of the MicroPython project, http://micropython.org/
#
# The MIT License (MIT)
#
# Copyright (c) 2022 Jim Mussared
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# This script compiles all packages in this repository (excluding unix-ffi)
# into a directory suitable for serving to "mip" via a static web server.

# Usage:
# ./tools/build.py --output /tmp/micropython-lib/v2

# The output directory (--output) will have the following layout
# /
#   index.json
#   file/
#     1d/
#       1dddc25d
#     c3/
#       c31d7eb7
#       c3a3934b
#     e3/
#       e39dbf64
#     ...
#   package/
#     6/                 <-- mpy version
#       aioble/
#         latest.json
#         0.1.json
#         ...
#       hmac/
#         latest.json
#         3.4.2-3.json
#         ...
#       pyjwt/
#         latest.json
#         0.1.json
#         ...
#     7/                 <-- other mpy versions
#       ...
#     py/                <-- "source" distribution
#       ...
#     ...

# index.json is:
# {
#   "v": 2,                      <-- file format version
#   "updated": <utc-seconds-since-1970>,
#   "packages": {
#     {
#       "name": "aioble",
#       "version": "0.1",        <-- Latest version of this package (always present, may be empty).
#       "author": "",            <-- Optional author (always present, may be empty).
#       "description": "...",    <-- Optional description (always present, may be empty).
#       "license": "MIT",        <-- SPDX short identifier (required).
#       "versions": {
#         "6": ["0.1", "0.2"],
#         "7": ["0.2", "0.3", "0.4"],
#         ...                    <-- Other bytecode versions
#         "py": ["0.1", "0.2", "0.3", "0.4"]
#       },
#       // The following entries were added in file format version 2.
#       path: "micropython/bluetooth/aioble",
#     },
#     ...
#   }
# }

# Each file in the "file" directory is the file contents (usually .mpy), named
# by the prefix of the sha256 hash of the contents. Files are never removed, and
# collisions are detected and will fail the compile, and the prefix length should
# be increased.
# As of September 2022, there are no collisions with a hash prefix length of 4,
# so the default of 8 should be sufficient for a while. Increasing the length
# doesn't invalidate old packages.

# Each package json (either latest.json or {version}.json) is:
# {
#   "v": 1,     <-- file format version
#   "hashes": [
#     ["aioble/server.mpy", "e39dbf64"],
#     ...
#   ],
#   "urls": [   <-- not used by micropython-lib packages
#     ["target/path.py", "http://url/foo/bar/path.py"],
#     ...
#   ],
#   "deps": [   <-- not used by micropython-lib packages
#     ["name", "version"],
#     ...
#   ]
#   "version": "0.1"
# }

# mip (or other tools) should request /package/{mpy_version}/{package_name}/{version}.json.

import glob
import hashlib
import json
import os
import shutil
import sys
import tempfile
import time


_JSON_VERSION_INDEX = 2
_JSON_VERSION_PACKAGE = 1


_COLOR_ERROR_ON = "\033[1;31m"
_COLOR_ERROR_OFF = "\033[0m"


# Create all directories in the path (such that the file can be created).
def ensure_path_exists(file_path):
    path = os.path.dirname(file_path)
    if not os.path.isdir(path):
        os.makedirs(path)


# Returns the sha256 of the specified file object.
def _get_file_hash(f):
    hs256 = hashlib.sha256()
    hs256.update(f.read())
    return hs256.hexdigest()


# Returns true if the two files contain identical contents.
def _identical_files(path_a, path_b):
    with open(path_a, "rb") as fa:
        with open(path_b, "rb") as fb:
            return fa.read() == fb.read()


# Helper to write the object as json to the specified path, creating any
# directories as required.
def _write_json(obj, path, minify=False):
    ensure_path_exists(path)
    with open(path, "w") as f:
        json.dump(
            obj, f, indent=(None if minify else 2), separators=((",", ":") if minify else None)
        )
        f.write("\n")


# Write the package json to package/{"py" or mpy_version}/{package}/{version}.json.
def _write_package_json(
    package_json, out_package_dir, mpy_version, package_name, version, replace
):
    path = os.path.join(out_package_dir, mpy_version, package_name, version + ".json")
    if replace or not os.path.exists(path):
        _write_json(package_json, path, minify=True)


# Format s with bold red.
def error_color(s):
    return _COLOR_ERROR_ON + s + _COLOR_ERROR_OFF


# Copy src to "file"/{short_hash[0:2]}/{short_hash}.
def _write_hashed_file(package_name, src, target_path, out_file_dir, hash_prefix_len):
    # Generate the full sha256 and the hash prefix to use as the output path.
    file_hash = _get_file_hash(src)
    short_file_hash = file_hash[:hash_prefix_len]
    # Group files into subdirectories using the first two bytes of the hash prefix.
    output_file = os.path.join(short_file_hash[:2], short_file_hash)
    output_file_path = os.path.join(out_file_dir, output_file)

    if os.path.exists(output_file_path):
        # If the file exists (e.g. from a previous run of this script), then ensure
        # that it's actually the same file.
        if not _identical_files(src.name, output_file_path):
            print(
                error_color("Hash collision processing:"),
                package_name,
                file=sys.stderr,
            )
            print("  File:        ", target_path, file=sys.stderr)
            print("  Short hash:  ", short_file_hash, file=sys.stderr)
            print("  Full hash:   ", file_hash, file=sys.stderr)
            with open(output_file_path, "rb") as f:
                print("  Target hash: ", _get_file_hash(f), file=sys.stderr)
            print("Try increasing --hash-prefix (currently {})".format(hash_prefix_len))
            sys.exit(1)
    else:
        # Create new file.
        ensure_path_exists(output_file_path)
        shutil.copyfile(src.name, output_file_path)

    return short_file_hash


# Convert the tagged .py file into a .mpy file and copy to the "file" output
# directory with it's hashed name. Updates the package_json with the file
# hash.
def _compile_as_mpy(
    package_name,
    package_json,
    tagged_path,
    target_path,
    opt,
    mpy_cross,
    mpy_cross_path,
    out_file_dir,
    hash_prefix_len,
):
    with tempfile.NamedTemporaryFile(mode="rb", suffix=".mpy", delete=True) as mpy_tempfile:
        try:
            mpy_cross.compile(
                tagged_path,
                dest=mpy_tempfile.name,
                src_path=target_path,
                opt=opt,
                mpy_cross=mpy_cross_path,
            )
        except mpy_cross.CrossCompileError as e:
            print(
                error_color("Error:"),
                "Unable to compile",
                target_path,
                "in package",
                package_name,
                file=sys.stderr,
            )
            print(e)
            sys.exit(1)

        short_mpy_hash = _write_hashed_file(
            package_name, mpy_tempfile, target_path, out_file_dir, hash_prefix_len
        )

        # Add the file to the package json.
        target_path_mpy = target_path[:-2] + "mpy"
        package_json["hashes"].append((target_path_mpy, short_mpy_hash))


# Copy the tagged .py file to the "file" output directory with it's hashed
# name. Updates the package_json with the file hash.
def _copy_as_py(
    package_name, package_json, tagged_path, target_path, out_file_dir, hash_prefix_len
):
    with open(tagged_path, "rb") as tagged_file:
        short_py_hash = _write_hashed_file(
            package_name, tagged_file, target_path, out_file_dir, hash_prefix_len
        )
    # Add the file to the package json.
    package_json["hashes"].append((target_path, short_py_hash))


# Update to the latest metadata, and add any new versions to the package in
# the index json.
def _update_index_package_metadata(index_package_json, metadata, mpy_version, package_path):
    index_package_json["version"] = metadata.version or ""
    index_package_json["author"] = ""  # TODO: Make manifestfile.py capture this.
    index_package_json["description"] = metadata.description or ""
    index_package_json["license"] = metadata.license or "MIT"
    if "versions" not in index_package_json:
        index_package_json["versions"] = {}
    if metadata.version:
        for v in ("py", mpy_version):
            if v not in index_package_json["versions"]:
                index_package_json["versions"][v] = []
            if metadata.version not in index_package_json["versions"][v]:
                print("  New version {}={}".format(v, metadata.version))
                index_package_json["versions"][v].append(metadata.version)

    # The following entries were added in file format version 2.
    index_package_json["path"] = package_path


def build(output_path, hash_prefix_len, mpy_cross_path):
    import manifestfile
    import mpy_cross

    out_file_dir = os.path.join(output_path, "file")
    out_package_dir = os.path.join(output_path, "package")

    path_vars = {
        "MPY_LIB_DIR": os.path.abspath(os.path.join(os.path.dirname(__file__), "..")),
    }

    index_json_path = os.path.join(output_path, "index.json")

    try:
        with open(index_json_path) as f:
            print("Updating existing index.json")
            index_json = json.load(f)
    except FileNotFoundError:
        print("Creating new index.json")
        index_json = {"packages": []}

    index_json["v"] = _JSON_VERSION_INDEX
    index_json["updated"] = int(time.time())

    # For now, don't process unix-ffi. In the future this can be extended to
    # allow a way to request unix-ffi packages via mip.
    lib_dirs = ["micropython", "python-stdlib", "python-ecosys"]

    mpy_version, _mpy_sub_version = mpy_cross.mpy_version(mpy_cross=mpy_cross_path)
    mpy_version = str(mpy_version)
    print("Generating bytecode version", mpy_version)

    for lib_dir in lib_dirs:
        for manifest_path in glob.glob(os.path.join(lib_dir, "**", "manifest.py"), recursive=True):
            package_path = os.path.dirname(manifest_path)
            print("{}".format(package_path))
            # .../foo/manifest.py -> foo
            package_name = os.path.basename(os.path.dirname(manifest_path))

            # Compile the manifest.
            manifest = manifestfile.ManifestFile(manifestfile.MODE_COMPILE, path_vars)
            manifest.execute(manifest_path)

            # Append this package to the index.
            if not manifest.metadata().version:
                print(error_color("Warning:"), package_name, "doesn't have a version.")

            # Try to find this package in the previous index.json.
            for p in index_json["packages"]:
                if p["name"] == package_name:
                    index_package_json = p
                    break
            else:
                print("  First-time package")
                index_package_json = {
                    "name": package_name,
                }
                index_json["packages"].append(index_package_json)

            _update_index_package_metadata(
                index_package_json, manifest.metadata(), mpy_version, package_path
            )

            # This is the package json that mip/mpremote downloads.
            mpy_package_json = {
                "v": _JSON_VERSION_PACKAGE,
                "hashes": [],
                "version": manifest.metadata().version or "",
            }
            py_package_json = {
                "v": _JSON_VERSION_PACKAGE,
                "hashes": [],
                "version": manifest.metadata().version or "",
            }

            for result in manifest.files():
                # This isn't allowed in micropython-lib anyway.
                if result.file_type != manifestfile.FILE_TYPE_LOCAL:
                    print(error_color("Error:"), "Non-local file not supported.", file=sys.stderr)
                    sys.exit(1)

                if not result.target_path.endswith(".py"):
                    print(
                        error_color("Error:"),
                        "Target path isn't a .py file:",
                        result.target_path,
                        file=sys.stderr,
                    )
                    sys.exit(1)

                # Tag each file with the package metadata and compile to .mpy
                # (and copy the .py directly).
                with manifestfile.tagged_py_file(result.full_path, result.metadata) as tagged_path:
                    _compile_as_mpy(
                        package_name,
                        mpy_package_json,
                        tagged_path,
                        result.target_path,
                        result.opt,
                        mpy_cross,
                        mpy_cross_path,
                        out_file_dir,
                        hash_prefix_len,
                    )
                    _copy_as_py(
                        package_name,
                        py_package_json,
                        tagged_path,
                        result.target_path,
                        out_file_dir,
                        hash_prefix_len,
                    )

            # Create/replace {package}/latest.json.
            _write_package_json(
                mpy_package_json,
                out_package_dir,
                mpy_version,
                package_name,
                "latest",
                replace=True,
            )
            _write_package_json(
                py_package_json, out_package_dir, "py", package_name, "latest", replace=True
            )

            # Write {package}/{version}.json, but only if it doesn't already
            # exist. A package version is "locked" the first time it's seen
            # by this script.
            if manifest.metadata().version:
                _write_package_json(
                    mpy_package_json,
                    out_package_dir,
                    mpy_version,
                    package_name,
                    manifest.metadata().version,
                    replace=False,
                )
                _write_package_json(
                    py_package_json,
                    out_package_dir,
                    "py",
                    package_name,
                    manifest.metadata().version,
                    replace=False,
                )

    # Write updated package index json, sorted by package name.
    index_json["packages"].sort(key=lambda p: p["name"])
    _write_json(index_json, index_json_path, minify=False)


def main():
    import argparse

    cmd_parser = argparse.ArgumentParser(description="Compile micropython-lib for serving to mip.")
    cmd_parser.add_argument("--output", required=True, help="output directory")
    cmd_parser.add_argument("--hash-prefix", default=8, type=int, help="hash prefix length")
    cmd_parser.add_argument("--mpy-cross", default=None, help="optional path to mpy-cross binary")
    cmd_parser.add_argument("--micropython", default=None, help="path to micropython repo")
    args = cmd_parser.parse_args()

    if args.micropython:
        sys.path.append(os.path.join(args.micropython, "tools"))  # for manifestfile
        sys.path.append(os.path.join(args.micropython, "mpy-cross"))  # for mpy_cross

    build(args.output, hash_prefix_len=max(4, args.hash_prefix), mpy_cross_path=args.mpy_cross)


if __name__ == "__main__":
    main()



# Test interruption of select.poll by EINTR signal, when
# MICROPY_PY_SELECT_POSIX_OPTIMISATIONS is enabled.

try:
    import time, gc, select, socket, _thread

    time.time_ns  # Check for time_ns on MicroPython
    select.poll  # Raises AttributeError for CPython implementations without poll()
except (ImportError, AttributeError):
    print("SKIP")
    raise SystemExit


def thread_main():
    lock.acquire()
    time.sleep(0.2)
    print("thread gc start")
    # The unix gc.collect() implementation will raise EINTR on other threads.
    # Could possibly use _thread._interrupt_main() instead if MicroPython had it.
    gc.collect()
    print("thread gc end")


# Start a thread to interrupt the main thread during its call to poll.
lock = _thread.allocate_lock()
lock.acquire()
_thread.start_new_thread(thread_main, ())

# Use a new UDP socket for tests, which should be writable but not readable.
s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
s.bind(socket.getaddrinfo("127.0.0.1", 8000)[0][-1])

# Create the poller object.
poller = select.poll()
poller.register(s, select.POLLIN)

# Poll on the UDP socket for a set timeout, which should be reached.
print("poll")
lock.release()
t0 = time.time_ns()
result = poller.poll(400)
dt_ms = (time.time_ns() - t0) / 1e6
print("result:", result)
if 380 <= dt_ms <= 600:
    print("dt in range")
else:
    print("dt not in range:", dt_ms)

# Clean up.
s.close()



"""
This is an auxiliary script that is used to compute valid PLL values to set
the CPU frequency to a given value.  The algorithm here appears as C code
for the machine.freq() function.
"""

from __future__ import print_function
import re


class MCU:
    def __init__(
        self, range_sysclk, range_m, range_n, range_p, range_q, range_vco_in, range_vco_out
    ):
        self.range_sysclk = range_sysclk
        self.range_m = range_m
        self.range_n = range_n
        self.range_p = range_p
        self.range_q = range_q
        self.range_vco_in = range_vco_in
        self.range_vco_out = range_vco_out


mcu_default = MCU(
    range_sysclk=range(2, 216 + 1, 2),
    range_m=range(2, 63 + 1),
    range_n=range(192, 432 + 1),
    range_p=range(2, 8 + 1, 2),
    range_q=range(2, 15 + 1),
    range_vco_in=range(1, 2 + 1),
    range_vco_out=range(192, 432 + 1),
)

mcu_table = {
    "stm32f413": MCU(
        range_sysclk=range(2, 100 + 1, 2),
        range_m=range(2, 63 + 1),
        range_n=range(50, 432 + 1),
        range_p=range(2, 8 + 1, 2),
        range_q=range(2, 15 + 1),
        range_vco_in=range(1, 2 + 1),
        range_vco_out=range(100, 432 + 1),
    ),
    "stm32h5": MCU(
        range_sysclk=range(2, 250 + 1, 2),
        range_m=range(1, 63 + 1),
        range_n=range(4, 512 + 1),
        range_p=range(2, 128 + 1, 2),
        range_q=range(1, 128 + 1),
        range_vco_in=range(1, 16 + 1),
        range_vco_out=range(150, 836 + 1),  # 150-420=medium, 192-836=wide
    ),
    "stm32h7": MCU(
        range_sysclk=range(2, 400 + 1, 2),  # above 400MHz currently unsupported
        range_m=range(1, 63 + 1),
        range_n=range(4, 512 + 1),
        range_p=range(2, 128 + 1, 2),
        range_q=range(1, 128 + 1),
        range_vco_in=range(1, 16 + 1),
        range_vco_out=range(150, 960 + 1),  # 150-420=medium, 192-960=wide
    ),
}


def close_int(x):
    return abs(x - round(x)) < 0.01


# original version that requires N/M to be an integer (for simplicity)
def compute_pll(hse, sys):
    for P in (2, 4, 6, 8):  # allowed values of P
        Q = sys * P / 48
        NbyM = sys * P / hse
        # N/M and Q must be integers
        if not (close_int(NbyM) and close_int(Q)):
            continue
        # VCO_OUT must be between 192MHz and 432MHz
        if not (192 <= hse * NbyM <= 432):
            continue
        # compute M
        M = int(192 // NbyM)
        while hse > 2 * M or NbyM * M < 192:
            M += 1
        # VCO_IN must be between 1MHz and 2MHz (2MHz recommended)
        if not (M <= hse):
            continue
        # compute N
        N = NbyM * M
        # N and Q are restricted
        if not (192 <= N <= 432 and 2 <= Q <= 15):
            continue
        # found valid values
        assert NbyM == N // M
        return (M, N, P, Q)
    # no valid values found
    return None


# improved version that doesn't require N/M to be an integer
def compute_pll2(hse, sys, relax_pll48):
    # Loop over the allowed values of P, looking for a valid PLL configuration
    # that gives the desired "sys" frequency.
    fallback = None
    for P in mcu.range_p:
        # VCO_OUT must be between 192MHz and 432MHz
        if sys * P not in mcu.range_vco_out:
            continue
        NbyM = float(sys * P) / hse  # float for Python 2
        # scan M
        M_min = mcu.range_n[0] // int(round(NbyM))  # starting value
        while mcu.range_vco_in[-1] * M_min < hse:
            M_min += 1
        # VCO_IN must be >=1MHz, but higher is better for stability so start high (low M)
        for M in range(M_min, hse + 1):
            # compute N
            N = NbyM * M
            # N must be an integer
            if not close_int(N):
                continue
            N = round(N)
            # N is restricted
            if N not in mcu.range_n:
                continue
            Q = float(sys * P) / 48  # float for Python 2
            # Q must be an integer in a set range
            if close_int(Q) and round(Q) in mcu.range_q:
                # found valid values
                return (M, N, P, Q)
            # Re-try Q to get at most 48MHz
            Q = (sys * P + 47) // 48
            if Q not in mcu.range_q:
                continue
            if fallback is None:
                # the values don't give 48MHz on PLL48 but are otherwise OK
                fallback = M, N, P, Q
    if relax_pll48:
        # might have found values which don't give 48MHz on PLL48
        return fallback
    else:
        # no valid values found which give 48MHz on PLL48
        return None


def compute_derived(hse, pll):
    hse = float(hse)  # float for Python 2
    M, N, P, Q = pll
    vco_in = hse / M
    vco_out = hse * N / M
    pllck = hse / M * N / P
    pll48ck = hse / M * N / Q
    return (vco_in, vco_out, pllck, pll48ck)


def verify_pll(hse, pll):
    M, N, P, Q = pll
    vco_in, vco_out, pllck, pll48ck = compute_derived(hse, pll)

    # verify ints
    assert close_int(M)
    assert close_int(N)
    assert close_int(P)
    assert close_int(Q)

    # verify range
    assert M in mcu.range_m
    assert N in mcu.range_n
    assert P in mcu.range_p
    assert Q in mcu.range_q
    assert mcu.range_vco_in[0] <= vco_in <= mcu.range_vco_in[-1]
    assert mcu.range_vco_out[0] <= vco_out <= mcu.range_vco_out[-1]


def compute_pll_table(source_clk, relax_pll48):
    valid_plls = []
    for sysclk in mcu.range_sysclk:
        pll = compute_pll2(source_clk, sysclk, relax_pll48)
        if pll is not None:
            verify_pll(source_clk, pll)
            valid_plls.append((sysclk, pll))
    return valid_plls


def generate_c_table(hse, valid_plls):
    valid_plls.sort()
    if (
        mcu.range_sysclk[-1] <= 0xFF
        and mcu.range_m[-1] <= 0x3F
        and mcu.range_p[-1] // 2 - 1 <= 0x3
    ):
        typedef = "uint16_t"
        sys_mask = 0xFF
        m_shift = 10
        m_mask = 0x3F
        p_shift = 8
        p_mask = 0x3
    else:
        typedef = "uint32_t"
        sys_mask = 0xFFFF
        m_shift = 24
        m_mask = 0xFF
        p_shift = 16
        p_mask = 0xFF
    print("#define PLL_FREQ_TABLE_SYS(pll) ((pll) & %d)" % (sys_mask,))
    print("#define PLL_FREQ_TABLE_M(pll) (((pll) >> %d) & %d)" % (m_shift, m_mask))
    print("#define PLL_FREQ_TABLE_P(pll) (((((pll) >> %d) & %d) + 1) * 2)" % (p_shift, p_mask))
    print("typedef %s pll_freq_table_t;" % (typedef,))
    print("// (M, P/2-1, SYS) values for %u MHz source" % hse)
    print("static const pll_freq_table_t pll_freq_table[%u] = {" % (len(valid_plls),))
    for sys, (M, N, P, Q) in valid_plls:
        print("    (%u << %u) | (%u << %u) | %u," % (M, m_shift, P // 2 - 1, p_shift, sys), end="")
        if M >= 2:
            vco_in, vco_out, pllck, pll48ck = compute_derived(hse, (M, N, P, Q))
            print(
                " // M=%u N=%u P=%u Q=%u vco_in=%.2f vco_out=%.2f pll48=%.2f"
                % (M, N, P, Q, vco_in, vco_out, pll48ck),
                end="",
            )
        print()
    print("};")


def print_table(hse, valid_plls):
    print("HSE =", hse, "MHz")
    print("sys :  M      N     P     Q : VCO_IN VCO_OUT   PLLCK PLL48CK")
    out_format = "%3u : %2u  %.1f  %.2f  %.2f :  %5.2f  %6.2f  %6.2f  %6.2f"
    for sys, pll in valid_plls:
        print(out_format % ((sys,) + pll + compute_derived(hse, pll)))
    print("found %u valid configurations" % len(valid_plls))


def search_header_for_hsx_values(filename, vals):
    regex_inc = re.compile(r'#include "(boards/[A-Za-z0-9_./]+)"')
    regex_def = re.compile(r"#define +(HSE_VALUE|HSI_VALUE) +\((\(uint32_t\))?([0-9]+)\)")
    with open(filename) as f:
        for line in f:
            line = line.strip()
            m = regex_inc.match(line)
            if m:
                # Search included file
                search_header_for_hsx_values(m.group(1), vals)
                continue
            m = regex_def.match(line)
            if m:
                # Found HSE_VALUE or HSI_VALUE
                val = int(m.group(3)) // 1000000
                if m.group(1) == "HSE_VALUE":
                    vals[0] = val
                else:
                    vals[1] = val
    return vals


def main():
    global mcu
    global out_format

    # parse input args
    import sys

    argv = sys.argv[1:]

    c_table = False
    mcu_series = "stm32f4"
    hse = None
    hsi = None

    while True:
        if argv[0] == "-c":
            c_table = True
            argv.pop(0)
        elif argv[0] == "-m":
            argv.pop(0)
            mcu_series = argv.pop(0).lower()
        else:
            break

    if len(argv) != 1:
        print("usage: pllvalues.py [-c] [-m <mcu_series>] <hse in MHz>")
        sys.exit(1)

    if argv[0].startswith("file:"):
        # extract HSE_VALUE, and optionally HSI_VALUE, from header file
        hse, hsi = search_header_for_hsx_values(argv[0][5:], [None, None])
        if hse is None:
            raise ValueError("%s does not contain a definition of HSE_VALUE" % argv[0])
    else:
        # HSE given directly as an integer
        hse = int(argv[0])

    # Select MCU parameters
    mcu = mcu_default
    for m in mcu_table:
        if mcu_series.startswith(m):
            mcu = mcu_table[m]
            break

    # Relax constraint on PLLQ being 48MHz on MCUs which have separate PLLs for 48MHz
    relax_pll48 = mcu_series.startswith(("stm32f413", "stm32f7", "stm32h5", "stm32h7"))

    hse_valid_plls = compute_pll_table(hse, relax_pll48)
    if hsi is not None:
        hsi_valid_plls = compute_pll_table(hsi, relax_pll48)

    if c_table:
        print("#if MICROPY_HW_CLK_USE_HSI")
        if hsi is not None:
            hsi_valid_plls.append((hsi, (0, 0, 2, 0)))
            generate_c_table(hsi, hsi_valid_plls)
        print("#else")
        if hsi is not None:
            hse_valid_plls.append((hsi, (0, 0, 2, 0)))
        hse_valid_plls.append((hse, (1, 0, 2, 0)))
        generate_c_table(hse, hse_valid_plls)
        print("#endif")
    else:
        print_table(hse, hse_valid_plls)


if __name__ == "__main__":
    main()



r"""Command-line tool to validate and pretty-print JSON

Usage::

    $ echo '{"json":"obj"}' | python -m json.tool
    {
        "json": "obj"
    }
    $ echo '{ 1.2:3.4}' | python -m json.tool
    Expecting property name enclosed in double quotes: line 1 column 3 (char 2)

"""
import sys
import json


def main():
    if len(sys.argv) == 1:
        infile = sys.stdin
        outfile = sys.stdout
    elif len(sys.argv) == 2:
        infile = open(sys.argv[1], "r")
        outfile = sys.stdout
    elif len(sys.argv) == 3:
        infile = open(sys.argv[1], "r")
        outfile = open(sys.argv[2], "w")
    else:
        raise SystemExit(sys.argv[0] + " [infile [outfile]]")
    with infile:
        try:
            obj = json.load(infile)
        except ValueError as e:
            raise SystemExit(e)
    with outfile:
        json.dump(obj, outfile, sort_keys=True, indent=4, separators=(",", ": "))
        outfile.write("\n")


if __name__ == "__main__":
    main()
